import functools
import weakref
from typing import Any, Callable


def weak_lru(maxsize: int = 128, typed: bool = False) -> Callable[..., Any]:
    """Decorator that allows safe use of lru_cache in class methods.

    Args:
        maxsize: maximum size. Defaults to 128.
        typed: typed. Defaults to False.

    Returns:
        wrapped method.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        ref = weakref.ref

        @functools.lru_cache(maxsize, typed)
        def _func(_self: Any, /, *args: Any, **kwargs: Any) -> Any:
            return func(_self(), *args, **kwargs)

        @functools.wraps(func)
        def wrapper(self: Any, /, *args: Any, **kwargs: Any) -> Any:
            return _func(ref(self), *args, **kwargs)

        wrapper.cache_clear = _func.cache_clear  # type: ignore[attr-defined]

        return wrapper

    return decorator



import time
from collections import defaultdict, deque
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Optional, Union

from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.ethercat.servo import EthercatServo
from ingenialink.exceptions import ILError
from ingenialink.pdo import PDOMap, RPDOMap, RPDOMapItem, TPDOMap, TPDOMapItem
from ingenialogger import get_logger

from ingeniamotion.enums import CommunicationType
from ingeniamotion.exceptions import IMError
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

logger = get_logger(__name__)


class PDOPoller:
    """Poll register values using PDOs."""

    def __init__(
        self,
        mc: "MotionController",
        servo: str,
        refresh_time: float,
        watchdog_timeout: Optional[float],
        buffer_size: int,
    ) -> None:
        """Constructor.

        Args:
            mc: MotionController instance
            servo: drive alias.
            refresh_time: PDO values refresh time.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.
            buffer_size: Maximum number of register readings to store.

        """
        super().__init__()
        self.__mc = mc
        self.__servo = servo
        self.__refresh_time = refresh_time
        self.__watchdog_timeout = watchdog_timeout
        self.__buffer_size = buffer_size
        self.__buffer: deque[tuple[float, list[Union[int, float, bytes]]]] = deque(
            maxlen=self.__buffer_size
        )
        self.__start_time: Optional[float] = None
        self.__tpdo_map: TPDOMap = TPDOMap()
        self.__rpdo_map: RPDOMap = RPDOMap()
        self.__fill_rpdo_map()
        self.__exception_callbacks: list[Callable[[ILError], None]] = []

    def start(self) -> None:
        """Start the poller."""
        self.__mc.capture.pdo.set_pdo_maps_to_slave(
            rpdo_maps=self.__rpdo_map, tpdo_maps=self.__tpdo_map, servo=self.__servo
        )
        self.__tpdo_map.subscribe_to_process_data_event(self._new_data_available)
        for callback in self.__exception_callbacks:
            self.__mc.capture.pdo.subscribe_to_exceptions(callback, servo=self.__servo)
        self.__start_time = time.time()
        self.__mc.capture.pdo.start_pdos(
            refresh_rate=self.__refresh_time,
            watchdog_timeout=self.__watchdog_timeout,
            servo=self.__servo,
        )

    def stop(self) -> None:
        """Stop the poller."""
        self.__mc.capture.pdo.stop_pdos(servo=self.__servo)
        self.__tpdo_map.unsubscribe_to_process_data_event()
        for callback in self.__exception_callbacks:
            self.__mc.capture.pdo.unsubscribe_to_exceptions(callback, servo=self.__servo)
        self.__mc.capture.pdo.remove_rpdo_map(servo=self.__servo, rpdo_map=self.__rpdo_map)
        self.__mc.capture.pdo.remove_tpdo_map(servo=self.__servo, tpdo_map=self.__tpdo_map)

    @property
    def data(self) -> tuple[list[float], list[list[Union[int, float, bytes]]]]:
        """Get the poller data. After the data is retrieved, the data buffers are cleared.

        Returns:
            A tuple with a list of the readings timestamps and a list of lists with
            the readings values.

        """
        time_stamps = []
        data: list[list[Union[int, float, bytes]]] = [[] for _ in range(len(self.__tpdo_map.items))]
        for _ in range(len(self.__buffer)):
            time_stamp, data_sample = self.__buffer.popleft()
            time_stamps.append(time_stamp)
            for item_index in range(len(self.__tpdo_map.items)):
                data[item_index].append(data_sample[item_index])

        return time_stamps, data

    def add_channels(self, registers: list[dict[str, Union[int, str]]]) -> None:
        """Configure the PDOs with the registers to be read.

        Args:
            registers : list of registers to add to the Poller.

        """
        self.__fill_tpdo_map(registers)

    def subscribe_to_exceptions(self, callback: Callable[[ILError], None]) -> None:
        """Get notified when an exception occurs on the PDO thread.

        Args:
            callback: Function to be called when an exception occurs.

        """
        self.__exception_callbacks.append(callback)

    def _new_data_available(self) -> None:
        """Add readings to the buffers.

        Raises:
            ValueError: If the poller has not been started yet.

        """
        if self.__start_time is None:
            raise ValueError("The poller has not been started yet.")
        time_stamp = round(time.time() - self.__start_time, 6)
        data_sample = [tpdo_map_item.value for tpdo_map_item in self.__tpdo_map.items]
        self.__buffer.append((time_stamp, data_sample))

    def __fill_rpdo_map(self) -> None:
        """Fill the RPDO Map with padding."""
        padding_rpdo_item = RPDOMapItem(size_bits=8)
        padding_rpdo_item.raw_data_bytes = int.to_bytes(0, 1, "little")
        self.__rpdo_map.add_item(padding_rpdo_item)

    def __fill_tpdo_map(self, registers: list[dict[str, Union[int, str]]]) -> None:
        """Fill the TPDO Map with the registers to be polled.

        Raises:
            ValueError: If there is a type mismatch when retrieving the register UID.
            ValueError: If there is a type mismatch when retrieving the register axis.

        """
        for register in registers:
            name = register.get("name", DEFAULT_SERVO)
            if not isinstance(name, str):
                raise ValueError(
                    f"Wrong type for the 'name' field. Expected 'str', got: {type(name)}"
                )
            axis = register.get("axis", DEFAULT_AXIS)
            if not isinstance(axis, int):
                raise ValueError(
                    f"Wrong type for the 'axis' field. Expected 'int', got: {type(axis)}"
                )
            tpdo_map_item = self.__mc.capture.pdo.create_pdo_item(
                register_uid=name, axis=axis, servo=self.__servo
            )
            self.__tpdo_map.add_item(tpdo_map_item)

    @property
    def available_samples(self) -> int:
        """Number of samples in the buffer."""
        return len(self.__buffer)


@dataclass
class PDONetwork:
    """Represents a PDO network."""

    alias: str
    network: EthercatNetwork
    refresh_rate: Optional[float]
    watchdog_timeout: Optional[float]

    __pdo_thread_status: bool = False

    def __pdo_thread_status_callback(self, status: bool) -> None:
        """Callback for PDO thread status changes.

        Args:
            status: The new status of the PDO thread.
        """
        self.__pdo_thread_status = status

    @property
    def is_active(self) -> bool:
        """Check if the PDO thread is active.

        Returns:
            True if the PDO thread is active. False otherwise.
        """
        return self.__pdo_thread_status

    def compare_configuration(
        self, refresh_rate: Optional[float], watchdog_timeout: Optional[float]
    ) -> bool:
        """Compares the current PDO thread configuration with another.

        Args:
            refresh_rate: Determines how often (seconds) the PDO values will be updated.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
                to the refresh rate.

        Returns:
            True if the configurations are equal, False otherwise.
        """
        return self.refresh_rate == refresh_rate and self.watchdog_timeout == watchdog_timeout

    @classmethod
    def create_and_subscribe(
        cls,
        alias: str,
        network: EthercatNetwork,
        refresh_rate: Optional[float] = None,
        watchdog_timeout: Optional[float] = None,
    ) -> "PDONetwork":
        """Creates a new PDONetwork instance and subscribes to EtherCAT network PDO thread status.

        Args:
            alias: The network alias.
            network: The EthercatNetwork instance.
            refresh_rate: Determines how often (seconds) the PDO values will be updated.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.

        Returns:
            A new PDONetwork instance.
        """
        net = cls(
            alias=alias,
            network=network,
            refresh_rate=refresh_rate,
            watchdog_timeout=watchdog_timeout,
        )
        net.network.subscribe_to_pdo_thread_status(callback=net.__pdo_thread_status_callback)
        return net

    def teardown(self) -> None:
        """Unsubscribes from network exceptions."""
        self.network.unsubscribe_from_pdo_thread_status(callback=self.__pdo_thread_status_callback)


class PDONetworksTracker:
    """Tracks the different networks that have been required to activate PDOs."""

    def __init__(self) -> None:
        self.__networks: dict[str, PDONetwork] = {}

    def is_network_tracked(self, alias: str) -> bool:
        """Checks if a network is being tracked.

        Args:
            alias: The network alias.

        Returns:
            True if the network is tracked, False otherwise.
        """
        return alias in self.__networks

    def add_network(
        self,
        alias: str,
        network: EthercatNetwork,
        refresh_rate: Optional[float] = None,
        watchdog_timeout: Optional[float] = None,
    ) -> None:
        """Add a network to the tracker.

        Args:
            alias: The network alias.
            network: The EthercatNetwork instance.
            refresh_rate: Determines how often (seconds) the PDO values will be updated.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.

        Raises:
            IMError: If the network is already tracked with a different configuration.
        """
        if not self.is_network_tracked(alias=alias):
            self.__networks[alias] = PDONetwork.create_and_subscribe(
                alias=alias,
                network=network,
                refresh_rate=refresh_rate,
                watchdog_timeout=watchdog_timeout,
            )
            network.activate_pdos(refresh_rate=refresh_rate, watchdog_timeout=watchdog_timeout)
            return
        if not self.__networks[alias].compare_configuration(
            refresh_rate=refresh_rate, watchdog_timeout=watchdog_timeout
        ):
            raise IMError("Inconsistent PDO thread configuration for already active network.")
        logger.warning("PDOs are already active for this network.")

    def remove_network(self, alias: str) -> None:
        """Remove a network from the tracker.

        Args:
            alias: The network alias.

        Raises:
            IMError: If the PDOs are not active for the network.
        """
        if not self.is_network_tracked(alias):
            raise IMError(f"PDOs are not active yet for network '{alias}'.")
        net = self.__networks[alias]
        # If the exception has happened when starting the PDOs, the PDOs are not active yet
        if not net.network.pdo_manager.is_active:
            logger.warning("PDOs are not active for this network.")
        else:
            net.network.deactivate_pdos()
        net.teardown()
        del self.__networks[alias]

    def get_il_network(self, alias: str) -> "EthercatNetwork":
        """Retrieves the Ethercat network for a specific alias.

        Args:
            alias: The network alias.

        Returns:
            The EthercatNetwork instance for the specified alias.

        Raises:
            IMError: If the PDOs are not active for the network.
        """
        if not self.is_network_tracked(alias):
            raise IMError(f"Cannot retrieve Ethercat network for '{alias}', it is not tracked.")
        return self.__networks[alias].network

    def is_active(self, alias: str) -> bool:
        """Checks if the PDO thread is active for a specific network.

        Args:
            alias: The network alias.

        Returns:
            bool: True if the PDO thread is active, False otherwise.
        """
        if self.is_network_tracked(alias):
            return self.__networks[alias].is_active
        return False


class PDONetworkManager:
    """Manage all the PDO functionalities.

    Args:
        mc: The MotionController.
    """

    def __init__(self, motion_controller: "MotionController") -> None:
        self.__mc = motion_controller

        self.__net_tracker = PDONetworksTracker()

        # Save the callbacks to add/remove, manage them when there is a reference to the network
        self.__send_process_data_add_callback: dict[str, list[Callable[[], None]]] = defaultdict(
            list
        )
        self.__receive_process_data_add_callback: dict[str, list[Callable[[], None]]] = defaultdict(
            list
        )
        self.__exception_add_callback: dict[str, list[Callable[[ILError], None]]] = defaultdict(
            list
        )
        self.__send_process_data_remove_callback: dict[str, list[Callable[[], None]]] = defaultdict(
            list
        )
        self.__receive_process_data_remove_callback: dict[str, list[Callable[[], None]]] = (
            defaultdict(list)
        )
        self.__exception_remove_callback: dict[str, list[Callable[[ILError], None]]] = defaultdict(
            list
        )

    def __evaluate_subscriptions(self, net: EthercatNetwork, alias: str) -> None:  # noqa: C901
        for servo, callbacks in self.__send_process_data_add_callback.items():
            if servo != alias:
                continue
            for callback in callbacks:
                net.pdo_manager.subscribe_to_send_process_data(callback)
        for servo, callbacks in self.__receive_process_data_add_callback.items():
            if servo != alias:
                continue
            for callback in callbacks:
                net.pdo_manager.subscribe_to_receive_process_data(callback)
        for servo, exception_callbacks in self.__exception_add_callback.items():
            if servo != alias:
                continue
            for exception_callback in exception_callbacks:
                net.pdo_manager.subscribe_to_exceptions(exception_callback)
        for servo, callbacks in self.__send_process_data_remove_callback.items():
            if servo != alias:
                continue
            for callback in callbacks:
                net.pdo_manager.unsubscribe_to_send_process_data(callback)
        for servo, callbacks in self.__receive_process_data_remove_callback.items():
            if servo != alias:
                continue
            for callback in callbacks:
                net.pdo_manager.unsubscribe_to_receive_process_data(callback)
        for servo, exception_callbacks in self.__exception_remove_callback.items():
            if servo != alias:
                continue
            for exception_callback in exception_callbacks:
                net.pdo_manager.unsubscribe_to_exceptions(exception_callback)

    def create_pdo_item(
        self,
        register_uid: str,
        axis: int = DEFAULT_AXIS,
        servo: str = DEFAULT_SERVO,
        value: Optional[Union[int, float]] = None,
    ) -> Union[RPDOMapItem, TPDOMapItem]:
        """Create a PDOMapItem by specifying a register UID.

        Args:
            register_uid: Register to be mapped.
            axis: servo axis. ``1`` by default.
            servo: servo alias to reference it. ``default`` by default.
            value: Initial value for an RPDO register.

        Returns:
            Mappable PDO item.

        Raises:
            ValueError: If there is a type mismatch retrieving the register object.
            AttributeError: If an initial value is not provided for an RPDO register.
        """
        drive = self.__mc._get_drive(servo=servo)
        return PDOMap.create_item_from_register_uid(
            uid=register_uid, axis=axis, dictionary=drive.dictionary, value=value
        )

    @staticmethod
    def create_pdo_maps(
        rpdo_map_items: Union[RPDOMapItem, list[RPDOMapItem]],
        tpdo_map_items: Union[TPDOMapItem, list[TPDOMapItem]],
    ) -> tuple[RPDOMap, TPDOMap]:
        """Create the RPDO and TPDO maps from PDOMapItems.

        Args:
            rpdo_map_items: The RPDOMapItems to be added to a RPDOMap.
            tpdo_map_items: The TDOMapItems to be added to a TPDOMap.

        Returns:
            RPDO and TPDO maps.

        """
        rpdo_map = RPDOMap.from_pdo_items(rpdo_map_items)
        tpdo_map = TPDOMap.from_pdo_items(tpdo_map_items)
        return rpdo_map, tpdo_map

    @staticmethod
    def create_empty_rpdo_map() -> RPDOMap:
        """Create an empty RPDOMap.

        Returns:
            The empty RPDOMap.

        """
        return RPDOMap()

    @staticmethod
    def create_empty_tpdo_map() -> TPDOMap:
        """Create an empty TPDOMap.

        Returns:
            The empty TPDOMap.

        """
        return TPDOMap()

    def set_pdo_maps_to_slave(
        self,
        rpdo_maps: Union[RPDOMap, list[RPDOMap]],
        tpdo_maps: Union[TPDOMap, list[TPDOMap]],
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Map the PDOMaps to the slave.

        Args:
            rpdo_maps: The RPDOMaps to be mapped.
            tpdo_maps: he TPDOMaps to be mapped.
            servo: servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If there is a type mismatch retrieving the drive object.
            ValueError: If not all elements of the RPDO map list are instances of a RPDO map.
            ValueError: If not all elements of the TPDO map list are instances of a TPDO map.
        """
        drive = self.__mc._get_drive(servo=servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")

        _rpdo_maps = [rpdo_maps] if isinstance(rpdo_maps, RPDOMap) else rpdo_maps
        _tpdo_maps = [tpdo_maps] if isinstance(tpdo_maps, TPDOMap) else tpdo_maps
        if not all(isinstance(rpdo_map, RPDOMap) for rpdo_map in _rpdo_maps):
            raise ValueError("Not all elements of the RPDO map list are instances of a RPDO map")
        if not all(isinstance(tpdo_map, TPDOMap) for tpdo_map in _tpdo_maps):
            raise ValueError("Not all elements of the TPDO map list are instances of a TPDO map")
        drive.set_pdo_map_to_slave(rpdo_maps=_rpdo_maps, tpdo_maps=_tpdo_maps)

    def clear_pdo_mapping(self, servo: str = DEFAULT_SERVO) -> None:
        """Clear the PDO mapping within the servo.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If there is a type mismatch retrieving the drive object.
        """
        drive = self.__mc._get_drive(servo=servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")
        drive.reset_pdo_mapping()

    def remove_rpdo_map(
        self,
        servo: str = DEFAULT_SERVO,
        rpdo_map: Optional[RPDOMap] = None,
        rpdo_map_index: Optional[int] = None,
    ) -> None:
        """Remove a RPDOMap from the RPDOMap list.

        The RPDOMap instance or the index of the map in the RPDOMap list
         should be provided.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            rpdo_map: The RPDOMap instance to be removed.
            rpdo_map_index: The index of the RPDOMap list to be removed.

        Raises:
            ValueError: If there is a type mismatch retrieving the drive object.
        """
        drive = self.__mc._get_drive(servo=servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")
        drive.remove_rpdo_map(rpdo_map=rpdo_map, rpdo_map_index=rpdo_map_index)

    def remove_tpdo_map(
        self,
        servo: str = DEFAULT_SERVO,
        tpdo_map: Optional[TPDOMap] = None,
        tpdo_map_index: Optional[int] = None,
    ) -> None:
        """Remove a TPDOMap from the TPDOMap list.

        The TPDOMap instance or the index of the map in the TPDOMap list
        should be provided.

        Args:
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
            tpdo_map: The TPDOMap instance to be removed.
            tpdo_map_index: The index of the TPDOMap list to be removed.

        Raises:
            ValueError: If there is a type mismatch retrieving the drive object.
        """
        drive = self.__mc._get_drive(servo=servo)
        if not isinstance(drive, EthercatServo):
            raise ValueError(f"Expected an EthercatServo. Got {type(drive)}")
        drive.remove_tpdo_map(tpdo_map=tpdo_map, tpdo_map_index=tpdo_map_index)

    def start_pdos(
        self,
        network_type: Optional[CommunicationType] = None,
        refresh_rate: Optional[float] = None,
        watchdog_timeout: Optional[float] = None,
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Start the PDO exchange process.

        Warning:
            Note that the PDO exchange will start for all servos connected to the same network.

        Args:
            network_type: Network type (EtherCAT or CANopen) on which to start the PDO exchange.
            refresh_rate: Determines how often (seconds) the PDO values will be updated.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                If `network_type` is provided, `servo` must be connected to that network.

        Raises:
            ValueError: If the MotionController is not connected to any Network.
            ValueError: If there is a type mismatch retrieving the network object.
        """
        if network_type in [None, CommunicationType.Ethercat]:
            if len(self.__mc.net) == 0:
                raise ValueError(
                    "No network created. Please create a network before starting PDOs."
                )
            # Start PDOs for the network the specified servo
            net = self.__mc._get_network(servo=servo)
        else:
            raise NotImplementedError

        if not isinstance(net, EthercatNetwork):
            raise ValueError(f"Expected EthercatNetwork. Got {type(net)}")

        self.__evaluate_subscriptions(net=net, alias=servo)
        self.__net_tracker.add_network(
            alias=self.__mc.servo_net[servo],
            network=net,
            refresh_rate=refresh_rate,
            watchdog_timeout=watchdog_timeout,
        )

    def stop_pdos(self, servo: str = DEFAULT_SERVO) -> None:
        """Stop the PDO exchange process.

        Warning:
            Note that the PDO exchange will stop for all servos connected to the same network.

        Args:
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                PDOs will be stopped in the network to which the servo is connected.

        Raises:
            IMError: If the PDOs are not active yet.
        """
        self.__net_tracker.remove_network(alias=self.__mc.servo_net[servo])

    def is_active(
        self, servo: Optional[str] = DEFAULT_SERVO, net_alias: Optional[str] = None
    ) -> bool:
        """Check if the PDO thread is active for the network to which the servo is connected.

        Alternatively, it can be checked for a specific network.
        If the network alias is provided, the servo will be ignored.

        Args:
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
            net_alias: network alias to reference it. ``None`` by default.

        Returns:
            True if the PDO thread is active. False otherwise.

        Raises:
            ValueError: If neither servo nor net_alias is provided.
        """
        if net_alias is not None:
            return self.__net_tracker.is_active(alias=net_alias)
        elif servo is not None:
            return self.__net_tracker.is_active(alias=self.__mc.servo_net[servo])
        raise ValueError("Either servo or net_alias must be provided.")

    def subscribe_to_send_process_data(
        self, callback: Callable[[], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Subscribe be notified when the RPDO values will be sent.

        Args:
            callback: Callback function.
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                The subscription will be added to the network to which the servo is connected.
        """
        net_alias = self.__mc.servo_net[servo]
        if self.__net_tracker.is_network_tracked(net_alias):
            net = self.__net_tracker.get_il_network(alias=net_alias)
            net.pdo_manager.subscribe_to_send_process_data(callback)
        else:
            self.__send_process_data_add_callback[servo].append(callback)

    def subscribe_to_receive_process_data(
        self, callback: Callable[[], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Subscribe be notified when the TPDO values are received.

        Args:
            callback: Callback function.
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                The subscription will be added to the network to which the servo is connected.
        """
        net_alias = self.__mc.servo_net[servo]
        if self.__net_tracker.is_network_tracked(net_alias):
            net = self.__net_tracker.get_il_network(alias=net_alias)
            net.pdo_manager.subscribe_to_receive_process_data(callback)
        else:
            self.__receive_process_data_add_callback[servo].append(callback)

    def subscribe_to_exceptions(
        self, callback: Callable[[ILError], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Subscribe be notified when there is an exception in the PDO process data thread.

        If a callback is subscribed, the PDO exchange process is paused when an exception is raised.
        It can be resumed using the `resume_pdos` method.

        Args:
            callback: Callback function.
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                The subscription will be added to the network to which the servo is connected.
        """
        net_alias = self.__mc.servo_net[servo]
        if self.__net_tracker.is_network_tracked(net_alias):
            net = self.__net_tracker.get_il_network(alias=net_alias)
            net.pdo_manager.subscribe_to_exceptions(callback)
        else:
            self.__exception_add_callback[servo].append(callback)

    def unsubscribe_to_send_process_data(
        self, callback: Callable[[], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Unsubscribe from the send process data notifications.

        Args:
            callback: Subscribed callback function.
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                The unsubscription will be removed from the network to which the servo is connected.

        Args:
            callback: Subscribed callback function.
        """
        net_alias = self.__mc.servo_net[servo]
        if self.__net_tracker.is_network_tracked(net_alias):
            net = self.__net_tracker.get_il_network(alias=net_alias)
            net.pdo_manager.unsubscribe_to_send_process_data(callback)
        else:
            self.__send_process_data_remove_callback[servo].append(callback)

    def unsubscribe_to_receive_process_data(
        self, callback: Callable[[], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Unsubscribe from the receive process data notifications.

        Args:
            callback: Subscribed callback function.
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                The unsubscription will be removed from the network to which the servo is connected.
        """
        net_alias = self.__mc.servo_net[servo]
        if self.__net_tracker.is_network_tracked(net_alias):
            net = self.__net_tracker.get_il_network(alias=net_alias)
            net.pdo_manager.unsubscribe_to_receive_process_data(callback)
        else:
            self.__receive_process_data_remove_callback[servo].append(callback)

    def create_poller(
        self,
        registers: list[dict[str, Union[int, str]]],
        servo: str = DEFAULT_SERVO,
        sampling_time: float = 0.125,
        buffer_size: int = 100,
        watchdog_timeout: Optional[float] = None,
        start: bool = True,
    ) -> PDOPoller:
        """Create a register Poller using PDOs.

        Args:
            registers : list of registers to add to the Poller.
                Dicts should have the follow format:

                .. code-block:: python

                    [
                        { # Poller register one
                            "name": "CL_POS_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        },
                        { # Poller register two
                            "name": "CL_VEL_FBK_VALUE",  # Register name.
                            "axis": 1  # Register axis.
                            # If it has no axis field, by default axis 1.
                        }
                    ]

            servo: servo alias to reference it. ``default`` by default.
            sampling_time: period of the sampling in seconds.
                By default ``0.125`` seconds.
            watchdog_timeout: The PDO watchdog time. If not provided it will be set proportional
             to the refresh rate.
            buffer_size: number maximum of sample for each data read.
                ``100`` by default.
            start: if ``True``, function starts poller, if ``False``
                poller should be started after. ``True`` by default.

        Returns:
            The poller instance.

        """
        poller = PDOPoller(
            mc=self.__mc,
            servo=servo,
            refresh_time=sampling_time,
            watchdog_timeout=watchdog_timeout,
            buffer_size=buffer_size,
        )
        poller.add_channels(registers)
        if start:
            poller.start()
        return poller

    def unsubscribe_to_exceptions(
        self, callback: Callable[[ILError], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Unsubscribe from the exceptions in the process data notifications.

        Args:
            callback: Subscribed callback function.
            servo: servo alias to reference it. ``DEFAULT_SERVO`` by default.
                The unsubscription will be removed from the network to which the servo is connected.
        """
        net_alias = self.__mc.servo_net[servo]
        if self.__net_tracker.is_network_tracked(net_alias):
            net = self.__net_tracker.get_il_network(alias=net_alias)
            net.pdo_manager.unsubscribe_to_exceptions(callback)
        else:
            self.__exception_remove_callback[servo].append(callback)



from . import enums
from .motion_controller import MotionController

try:
    from ._version import __version__  # noqa: F401
except ModuleNotFoundError:
    __version__ = "development"

__all__ = ["__version__", "MotionController", "enums"]



import warnings
from typing import Any


class IMError(Exception):
    """Exceptions raised by IngeniaMotion."""


class IMMonitoringError(IMError):
    """Monitoring error raised by IngeniaMotion."""


class IMDisturbanceError(IMError):
    """Disturbance error raised by IngeniaMotion."""


class IMStatusWordError(IMError):
    """Status word error raised by IngeniaMotion."""


class IMRegisterNotExistError(IMError):
    """Error raised by IngeniaMotion when a register not exists."""


class IMRegisterWrongAccessError(IMError):
    """Error raised by IngeniaMotion when trying to write to a read-only register."""


class IMTimeoutError(IMError):
    """Error raised by IngeniaMotion when a timeout has occurred."""


class IMFirmwareLoadError(IMError):
    """Error raised by IngeniaMotion when a firmware file could not be loaded."""


# WARNING: Deprecated aliases
_DEPRECATED = {
    "IMException": "IMError",
    "IMRegisterNotExist": "IMRegisterNotExistError",
    "IMRegisterWrongAccess": "IMRegisterWrongAccessError",
}


def __getattr__(name: str) -> Any:
    if name in _DEPRECATED:
        warnings.warn(
            f"{name} is deprecated, use {_DEPRECATED[name]} instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[_DEPRECATED[name]]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Callable, Optional, cast

import ingenialogger
from ingenialink.ethercat.servo import EthercatServo

from ingeniamotion.enums import FSoEState
from ingeniamotion.fsoe_master.fsoe import FSOE_MASTER_INSTALLED
from ingeniamotion.metaclass import DEFAULT_SERVO

if FSOE_MASTER_INSTALLED:
    from ingeniamotion.fsoe_master.handler import FSoEMasterHandler

if TYPE_CHECKING:
    from ingenialink.ethercat.network import EthercatNetwork

    from ingeniamotion.motion_controller import MotionController

__all__ = ["FSOE_MASTER_INSTALLED", "FSoEMaster", "FSoEError"]


@dataclass
class FSoEError:
    """FSoE Error descriptor."""

    servo: str
    transition_name: str
    description: str


class FSoEMaster:
    """FSoE Master.

    Args:
        motion_controller: The MotionController instance.

    """

    def __init__(self, motion_controller: "MotionController") -> None:
        self.logger = ingenialogger.get_logger(__name__)
        self.__mc = motion_controller
        self._handlers: dict[str, FSoEMasterHandler] = {}
        self.__next_connection_id = 1
        self._error_observers: list[Callable[[FSoEError], None]] = []

    def create_fsoe_master_handler(
        self,
        use_sra: bool,
        servo: str = DEFAULT_SERVO,
        fsoe_master_watchdog_timeout: Optional[float] = None,
        state_change_callback: Optional[Callable[[FSoEState], None]] = None,
    ) -> "FSoEMasterHandler":
        """Create an FSoE Master handler linked to a Safe servo drive.

        Raises:
            TypeError: If the servo is not an EthercatServo.

        Args:
            use_sra: True to use SRA, False otherwise.
            servo: servo alias to reference it. ``default`` by default.
            fsoe_master_watchdog_timeout: The FSoE master watchdog timeout in seconds.
            state_change_callback: Optional callback to be called when the
                FSoE master handler state changes.

        Returns:
            An instance of FSoEMasterHandler.
        """
        if fsoe_master_watchdog_timeout is None:
            fsoe_master_watchdog_timeout = FSoEMasterHandler.DEFAULT_WATCHDOG_TIMEOUT_S
        node = self.__mc.servos[servo]
        if not isinstance(node, EthercatServo):
            raise TypeError("Functional Safety over Ethercat is only available for Ethercat servos")
        net = cast("EthercatNetwork", self.__mc._get_network(servo=servo))

        master_handler = FSoEMasterHandler(
            servo=node,
            net=net,
            use_sra=use_sra,
            connection_id=self.__next_connection_id,
            watchdog_timeout=fsoe_master_watchdog_timeout,
            report_error_callback=partial(self._notify_errors, servo=servo),
            state_change_callback=state_change_callback,
        )
        self._handlers[servo] = master_handler
        self.__next_connection_id += 1
        return master_handler

    def start_master(self, start_pdos: bool = False) -> None:
        """Start an FSoE Master handler.

        Args:
            start_pdos: if ``True``, start the PDO exchange, if ``False``
                the PDO exchange should be started after. ``False`` by default.
        """
        for master_handler in self._handlers.values():
            master_handler.start()

        if start_pdos:
            for servo in self._handlers:
                self.__mc.capture.pdo.start_pdos(servo=servo)

    def configure_pdos(self, start_pdos: bool = False, start_master: bool = False) -> None:
        """Configure the PDOs used for the Safety PDUs.

        Args:
            start_pdos: if ``True``, start the PDO exchange, if ``False``
                the PDO exchange should be started after. ``False`` by default.
            start_master: if ``True``, start the FSoE master handlers after
                configuring the PDOs. ``False`` by default.
        """
        self._configure_and_set_pdo_maps_to_slaves()
        if start_master:
            self.start_master(start_pdos=start_pdos)
        elif start_pdos:
            for servo in self._handlers:
                self.__mc.capture.pdo.start_pdos(servo=servo)

    def stop_master(self, stop_pdos: bool = False) -> None:
        """Stop all the FSoE Master handlers.

        Args:
            stop_pdos: if ``True``, stop the PDO exchange. ``False`` by default.

        """
        for master_handler in self._handlers.values():
            if master_handler.running:
                master_handler.stop()
        if stop_pdos:
            for servo in self._handlers:
                self.__mc.capture.pdo.stop_pdos(servo=servo)

    def sto_deactivate(self, servo: str = DEFAULT_SERVO) -> None:
        """Deactivate the Safety Torque Off.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.sto_deactivate()

    def sto_activate(self, servo: str = DEFAULT_SERVO) -> None:
        """Activate the Safety Torque Off.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.sto_activate()

    def ss1_deactivate(self, servo: str = DEFAULT_SERVO) -> None:
        """Deactivate the SS1.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.ss1_deactivate()

    def ss1_activate(self, servo: str = DEFAULT_SERVO) -> None:
        """Activate the SS1.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.ss1_activate()

    def sout_disable(self, servo: str = DEFAULT_SERVO) -> None:
        """Deactivate the Safety Output.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.sout_disable()

    def sout_enable(self, servo: str = DEFAULT_SERVO) -> None:
        """Activate the Safety Output.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.sout_enable()

    def get_safety_inputs_value(self, servo: str = DEFAULT_SERVO) -> bool:
        """Get a drive's safe inputs register value.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
           The safe inputs value.

        """
        master_handler = self._handlers[servo]
        return master_handler.safe_inputs_value()

    def get_safety_address(self, servo: str = DEFAULT_SERVO) -> int:
        """Get the drive's FSoE slave address.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            The FSoE slave address.

        """
        master_handler = self._handlers[servo]
        return master_handler.get_safety_address()

    def set_safety_address(self, address: int, servo: str = DEFAULT_SERVO) -> None:
        """Set the drive's FSoE slave address.

        Args:
            address: The address to be set.
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        return master_handler.set_safety_address(address)

    def check_sto_active(self, servo: str = DEFAULT_SERVO) -> bool:
        """Check if the STO is active in a given servo.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            True if the STO is active. False otherwise.

        """
        master_handler = self._handlers[servo]
        return master_handler.is_sto_active()

    def wait_for_state_data(
        self, servo: str = DEFAULT_SERVO, timeout: Optional[float] = None
    ) -> None:
        """Wait for an FSoE master handler to reach the Data state.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            timeout : how many seconds to wait for the FSoE master to reach the
                Data state, if ``None`` it will wait forever.
                ``None`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.wait_for_data_state(timeout)

    def set_fail_safe(self, fail_safe: bool, servo: str = DEFAULT_SERVO) -> None:
        """Set the fail-safe mode of the FSoE master handler.

        Args:
            fail_safe: True to set the fail-safe mode, False to remove it.
            servo: servo alias to reference it. ``default`` by default.

        """
        master_handler = self._handlers[servo]
        master_handler.set_fail_safe(fail_safe)

    def get_fsoe_master_state(self, servo: str = DEFAULT_SERVO) -> FSoEState:
        """Get the servo's FSoE master handler state.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            The servo's FSoE master handler state.

        """
        master_handler = self._handlers[servo]
        return master_handler.state

    def subscribe_to_errors(self, callback: Callable[[FSoEError], None]) -> None:
        """Subscribe to the FSoE errors.

        Args:
            callback: Subscribed callback function.

        """
        if callback in self._error_observers:
            return
        self._error_observers.append(callback)

    def unsubscribe_from_errors(self, callback: Callable[[FSoEError], None]) -> None:
        """Unsubscribe from the FSoE errors.

        Args:
            callback: Subscribed callback function.

        """
        if callback not in self._error_observers:
            return
        self._error_observers.remove(callback)

    def _notify_errors(self, transition_name: str, error_description: str, servo: str) -> None:
        """Notify subscribers when an FSoE error occurs.

        Args:
            transition_name: FSoE transition name.
            error_description: FSoE error description.
            servo: The servo alias.

        """
        for callback in self._error_observers:
            callback(FSoEError(servo, transition_name, error_description))

    def _delete_master_handler(self, servo: str = DEFAULT_SERVO) -> None:
        """Delete the master handler instance.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        """
        if servo not in self._handlers:
            return
        self._handlers[servo].delete()
        del self._handlers[servo]

    def _configure_and_set_pdo_maps_to_slaves(self) -> None:
        """Configure the PDOMaps used by the Safety PDUs in the slaves."""
        for master_handler in self._handlers.values():
            master_handler.configure_pdo_maps()
            master_handler.set_pdo_maps_to_slave()

    def _remove_pdo_maps_from_slaves(self) -> None:
        """Remove the PDOMaps used by the Safety PDUs from the slaves."""
        for master_handler in self._handlers.values():
            master_handler.remove_pdo_maps_from_slave()



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ingeniamotion import MotionController
from ingeniamotion.enums import SensorType
from ingeniamotion.wizard_tests.feedbacks_tests.feedback_test import Feedbacks


class DigitalIncremental2Test(Feedbacks):
    """Digital incremental 2 test class."""

    BACKUP_REGISTERS_QEI2: list[str] = ["FBK_DIGENC2_POLARITY"]

    FEEDBACK_POLARITY_REGISTER = "FBK_DIGENC2_POLARITY"

    SENSOR_TYPE_FEEDBACK_TEST = SensorType.QEI2

    def __init__(
        self, mc: "MotionController", servo: str, axis: int, logger_drive_name: Optional[str] = None
    ) -> None:
        super().__init__(mc, servo, axis, logger_drive_name)
        self.backup_registers_names.extend(self.BACKUP_REGISTERS_QEI2)



import json
import platform
import tempfile
import time
import zipfile
from collections import OrderedDict
from dataclasses import dataclass
from functools import partial
from os import path
from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast

import ifaddr
import ingenialogger
from ingenialink import CanBaudrate, CanDevice, NetDevEvt, NetState
from ingenialink.canopen.network import CAN_CHANNELS, CanopenNetwork
from ingenialink.canopen.servo import CanopenServo
from ingenialink.dictionary import Interface
from ingenialink.emcy import EmergencyMessage
from ingenialink.enums.register import RegAccess, RegDtype
from ingenialink.enums.servo import ServoState
from ingenialink.eoe.network import EoENetwork
from ingenialink.ethercat.network import EthercatNetwork, GilReleaseConfig
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.exceptions import ILError
from ingenialink.network import SlaveInfo
from ingenialink.register import Register
from ingenialink.servo import DictionaryFactory, Servo
from ingenialink.virtual.network import VirtualNetwork
from ping3 import ping
from virtual_drive.core import VirtualDrive

from ingeniamotion.exceptions import IMFirmwareLoadError, IMRegisterWrongAccessError

if TYPE_CHECKING:
    from ingenialink.ethercat.servo import EthercatServo
    from ingenialink.ethernet.servo import EthernetServo
    from ingenialink.virtual.servo import VirtualServo

    from ingeniamotion.motion_controller import MotionController

import contextlib

from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO

RUNNING_ON_WINDOWS = platform.system() == "Windows"

if RUNNING_ON_WINDOWS:
    pass

FILE_EXT_SFU = ".sfu"
FILE_EXT_LFU = ".lfu"
FIRMWARE_FILE_FAIL_MSG = "The firmware file could not be loaded correctly"

logger = ingenialogger.get_logger(__name__)


@dataclass
class IMRegisterUpdateObserver:
    """Ingeniamotion register update observer."""

    im_callback: Callable[[str, Servo, Register, Union[int, float, str, bytes]], None]
    alias: str


@dataclass
class IMEmergencyMessageObserver:
    """Ingeniamotion emergency message observer."""

    im_callback: Callable[[str, EmergencyMessage], None]
    alias: str


class Communication:
    """Communication."""

    FORCE_SYSTEM_BOOT_COCO_REGISTER = "DRV_BOOT_COCO_FORCE"
    FORCE_SYSTEM_BOOT_MOCO_REGISTER = "DRV_BOOT_MOCO_MODE"
    SYSTEM_RESET_MOCO_REGISTER = "DRV_BOOT_RESET"

    PASSWORD_FORCE_BOOT_COCO = 0x424F4F54
    PASSWORD_FORCE_BOOT_MOCO = 0x426F6F74
    PASSWORD_SYSTEM_RESET = 0x72657365

    ENSEMBLE_FIRMWARE_EXTENSION = ".zfu"
    ENSEMBLE_SLAVE_REV_NUM_MASK = 0x1000000

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.logger = logger
        self.__virtual_drive: Optional[VirtualDrive] = None
        self.register_update_observers: dict[Servo, list[IMRegisterUpdateObserver]] = {}
        self.emergency_messages_observers: dict[Servo, list[IMEmergencyMessageObserver]] = {}

    def __disconnect_callback(self, servo: Servo) -> None:
        alias = None
        for servo_alias, servo in self.mc.servos.items():
            if servo.target == servo.target:
                alias = servo_alias
                break
        if alias is None:
            raise ValueError("Servo not found in the communication controller.")

        network = self.mc._get_network(alias)
        if isinstance(network, VirtualNetwork) and self.__virtual_drive:
            self.__virtual_drive.stop()
            self.__virtual_drive = None
        del self.mc.servos[alias]
        net_name = self.mc.servo_net.pop(alias)
        servo_count = list(self.mc.servo_net.values()).count(net_name)
        if self.mc.fsoe_is_installed:
            self.mc.fsoe._delete_master_handler(alias)
        if servo_count == 0:
            del self.mc.net[net_name]

    def connect_servo_eoe(
        self,
        ip: str,
        dict_path: Optional[str] = None,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[EthernetNetwork, "EthernetServo"]:
        """Connect to target servo by Ethernet over EtherCAT.

        Args:
            ip : servo IP.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The Ethernet network and the connected servo.

        Raises:
            TypeError: If the dict_path argument is missing.
            FileNotFoundError: If the dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        if not dict_path:
            raise TypeError("dict_path argument is missing")
        return self.__servo_connect(
            ip,
            dict_path,
            alias,
            port,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
            is_eoe=True,
        )

    def connect_servo_ethernet(
        self,
        ip: str,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[EthernetNetwork, "EthernetServo"]:
        """Connect to target servo by Ethernet.

        Args:
            ip : servo IP
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            connection_timeout: Timeout in seconds for connection.
                ``1`` seconds by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The Ethernet network and the connected servo.

        Raises:
            FileNotFoundError: If the dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")
        return self.__servo_connect(
            ip,
            dict_path,
            alias,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
        )

    def connect_servo_virtual(
        self,
        dict_path: Optional[str] = None,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[VirtualNetwork, "VirtualServo"]:
        """Connect to the virtual drive using an ethernet communication.

        Args:
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            connection_timeout: Timeout in seconds for connection.
                ``1`` seconds by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The virtual network and the connected servo.

        Raises:
            FileNotFoundError: If the dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        if dict_path is not None and not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")

        if self.__virtual_drive is None:
            self.__virtual_drive = VirtualDrive(port, dictionary_path=dict_path)
            self.__virtual_drive.start()

        net = VirtualNetwork()
        self.mc.net[alias] = net
        servo = net.connect_to_slave(
            self.__virtual_drive.dictionary_path,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
            disconnect_callback=self.__disconnect_callback,
        )

        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = alias
        return net, servo

    def __servo_connect(
        self,
        ip: str,
        dict_path: str,
        alias: str,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
        is_eoe: bool = False,
    ) -> tuple[EthernetNetwork, "EthernetServo"]:
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")

        net = EthernetNetwork()
        self.mc.net[alias] = net
        servo = net.connect_to_slave(
            ip,
            dict_path,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
            is_eoe=is_eoe,
            disconnect_callback=self.__disconnect_callback,
        )

        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = alias
        return net, servo

    def connect_servo_eoe_service(
        self,
        ifname: str,
        dict_path: str,
        ip: str = "192.168.3.22",
        slave: int = 1,
        port: int = 1061,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[EoENetwork, "EthernetServo"]:
        r"""Connect to target servo by Ethernet over EtherCAT.

        Args:
            ifname : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            dict_path : servo dictionary path.
            ip : IP address to be assigned to the servo.
            slave : slave index. ``1`` by default.
            port : servo port. ``1061`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The EoE network and the connected servo.

        Raises:
            NotImplementedError: If this method is run in Linux.
            FileNotFoundError: If the dict file doesn't exist.
            ValueError: ip must be a subnetwork of 192.168.3.0/24
            ingenialink.exceptions.ILError: If the EoE service is not running
            ingenialink.exceptions.ILError: If the EoE service cannot be started on the network
                                            interface.
        """
        if not RUNNING_ON_WINDOWS:
            raise NotImplementedError("EoE service only works on windows.")
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")
        if ifname not in self.mc.net:
            self.mc.net[ifname] = EoENetwork(ifname)
        net = cast("EoENetwork", self.mc.net[ifname])
        try:
            servo = net.connect_to_slave(
                slave,
                ip,
                dict_path,
                port,
                servo_status_listener=servo_status_listener,
                net_status_listener=net_status_listener,
                disconnect_callback=self.__disconnect_callback,
            )
        except ILError as e:
            if len(net.servos) == 0:
                del self.mc.net[ifname]
            raise e
        servo.slave = slave  # type: ignore [attr-defined]
        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = ifname
        return net, servo

    def connect_servo_eoe_service_interface_ip(
        self,
        interface_ip: str,
        dict_path: str,
        ip: str = "192.168.3.22",
        slave: int = 1,
        port: int = 1061,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[EoENetwork, "EthernetServo"]:
        """Connect to target servo by Ethernet over EtherCAT.

        Args:
            interface_ip : IP of the interface to be connected to.
            dict_path : servo dictionary path.
            ip : IP address to be assigned to the servo.
            slave : slave index. ``1`` by default.
            port : servo port. ``1061`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The EoE network and the connected servo.

        Raises:
            TypeError: If the dict_path argument is missing.
            IndexError: If interface index is out of range.
            FileNotFoundError: If the dict file doesn't exist.
            ValueError: ip must be a subnetwork of 192.168.3.0/24
            ingenialink.exceptions.ILError: If the EoE service is not running
            ingenialink.exceptions.ILError: If the EoE service cannot be started on the network
                                            interface.
        """
        return self.connect_servo_eoe_service(
            self.get_ifname_from_interface_ip(interface_ip),
            dict_path,
            ip,
            slave,
            port,
            alias,
            servo_status_listener,
            net_status_listener,
        )

    def connect_servo_comkit(
        self,
        ip: str,
        coco_dict_path: str,
        moco_dict_path: str,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[EthernetNetwork, "EthernetServo"]:
        """Connect to target servo using a COM-KIT.

        Args:
            ip : servo IP
            coco_dict_path : COCO dictionary path.
            moco_dict_path : MOCO dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            connection_timeout: Timeout in seconds for connection.
                ``1`` seconds by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The Ethernet network and the connected servo.

        Raises:
            FileNotFoundError: If a dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        for dict_path in [coco_dict_path, moco_dict_path]:
            if not path.isfile(dict_path):
                raise FileNotFoundError(f"{dict_path} file does not exist!")
        net, servo = self.__servo_connect(
            ip,
            moco_dict_path,
            alias,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
        )
        coco_dict = DictionaryFactory.create_dictionary(coco_dict_path, Interface.ETH)
        self.mc.servos[alias].dictionary += coco_dict
        return net, servo

    @classmethod
    def __get_adapter_name(cls, index: int) -> str:
        """Returns the adapter name of an adapter based on its index.

        Args:
            index : position of interface selected in
                :func:`get_interface_name_list`.
        """
        adapter_name = cls.get_interface_name_list()[index]
        adapter_guid = cls.get_network_adapters()[adapter_name]
        if RUNNING_ON_WINDOWS:
            return f"\\Device\\NPF_{adapter_guid}"
        return adapter_name

    def get_ifname_from_interface_ip(self, address: str) -> str:
        """Returns interface name based on the address ip of an interface.

        Args:
            address : ip expected adapter is expected to
            be configured with.

        Raises:
            ValueError: In case the input is not valid or the adapter
            is not found.

        Returns:
            Ifname of the controller.
        """
        try:
            index = self._get_interface_index_by_address(address)
        except IndexError:
            raise ValueError(
                f"Could not found a adapter configured as {address} to connect as EtherCAT master"
            )
        return self.__get_adapter_name(index)

    @staticmethod
    def _get_interface_index_by_address(address: str) -> int:
        """Get the interface index by searching its IP address.

        Args:
            address: IP address of the adapter.

        Returns:
            Interface's index.

        Raises:
            IndexError: If no adapter matches the IP address.

        """
        for idx, adapter in enumerate(ifaddr.get_adapters()):
            for ip in adapter.ips:
                if ip.is_IPv4 and ip.ip == address:
                    return idx
        raise IndexError

    def get_ifname_by_index(self, index: int) -> str:
        """Return interface name by index.

        Args:
            index : position of interface selected in
                :func:`get_interface_name_list`.

        Returns:
            Real name of selected interface.
            It can be used for functions :func:`connect_servo_eoe_service` and
            :func:`connect_servo_ethercat`.

        Raises:
            IndexError: If interface index is out of range.

        """
        return self.__get_adapter_name(index)

    @classmethod
    def get_interface_name_list(cls) -> list[str]:
        """Get interface list.

        Returns:
            List with interface readable names.

        """
        network_adapters = cls.get_network_adapters()
        return list(network_adapters)

    @staticmethod
    def get_network_adapters() -> dict[str, str]:
        """Get the detected network adapters.

        Returns:
            Dictionary with interface readable names as keys and GUIDs as values.

        """
        adapters_dict = {}
        for adapter in ifaddr.get_adapters():
            guid = adapter.name.decode() if isinstance(adapter.name, bytes) else adapter.name
            if guid in adapters_dict:
                continue
            adapters_dict[guid] = adapter.nice_name
        if not RUNNING_ON_WINDOWS:
            return {name: guid for guid, name in adapters_dict.items()}
        if not EthercatNetwork.pysoem_available():
            logger.warning("npcap not available, network adapters list will not be complete")
            return {name: guid for guid, name in adapters_dict.items()}
        for index, guid, name in EthercatNetwork.find_adapters():
            if guid in adapters_dict:
                continue
            adapters_dict[guid] = name
        return {name: guid for guid, name in adapters_dict.items()}

    def get_available_canopen_devices(self) -> dict[CanDevice, list[int]]:
        """Return the list of available CAN devices (those connected and with drivers installed).

        Returns:
            Dict of available CAN devices and channels. For example:
            {
                CanDevice.KVASER: [0, 1]
                CanDevice.PCAN: [0]
            }
        """
        available_devices: dict[CanDevice, list[int]] = {}
        can_net = None
        for net_key in self.mc.net:
            net = self.mc.net[net_key]
            if isinstance(net, CanopenNetwork):
                can_net = net
                break
        if can_net is None:
            can_net = CanopenNetwork(CanDevice.KVASER)
        for device, channel in can_net.get_available_devices():
            can_device = CanDevice(device)
            can_channel = CAN_CHANNELS[device].index(channel)
            if can_device not in available_devices:
                available_devices[can_device] = [can_channel]
            else:
                available_devices[can_device].append(can_channel)
        return available_devices

    def connect_servo_eoe_service_interface_index(
        self,
        if_index: int,
        dict_path: str,
        ip: str = "192.168.3.22",
        slave: int = 1,
        port: int = 1061,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[EoENetwork, "EthernetServo"]:
        """Connect to target servo by Ethernet over EtherCAT.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
            dict_path : servo dictionary path.
            ip : IP address to be assigned to the servo.
            slave : slave index. ``1`` by default.
            port : servo port. ``1061`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The EoE network and the connected servo.

        Raises:
            TypeError: If the dict_path argument is missing.
            IndexError: If interface index is out of range.
            FileNotFoundError: If the dict file doesn't exist.
            ValueError: ip must be a subnetwork of 192.168.3.0/24
            ingenialink.exceptions.ILError: If the EoE service is not running
            ingenialink.exceptions.ILError: If the EoE service cannot be started on the network
                                            interface.
        """
        return self.connect_servo_eoe_service(
            self.get_ifname_by_index(if_index),
            dict_path,
            ip,
            slave,
            port,
            alias,
            servo_status_listener,
            net_status_listener,
        )

    def scan_servos_eoe_service(self, ifname: str) -> list[int]:
        r"""Return a List of available servos.

        Args:
            ifname : interface name. It should have format
                ``\\Device\\NPF_[...]``.

        Returns:
            Drives available in the target interface.

        Raises:
            NotImplementedError: If this method is run in Linux.
            ingenialink.exceptions.ILError: If the EoE service is not running
            TypeError: If some parameter has a wrong type.
        """
        if not RUNNING_ON_WINDOWS:
            raise NotImplementedError("EoE service only works on windows.")
        net = self.mc.net[ifname] if ifname in self.mc.net else EoENetwork(ifname)
        slaves = net.scan_slaves()
        if not isinstance(slaves, list):
            raise TypeError("Slaves are not saved in a list")
        return slaves

    def scan_servos_eoe_service_interface_index(self, if_index: int) -> list[int]:
        """Return a list of available servos.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.

        Returns:
            Drives available in the target interface.

        Raises:
            IndexError: If interface index is out of range.
            ingenialink.exceptions.ILError: If the EoE service is not running

        """
        return self.scan_servos_eoe_service(self.get_ifname_by_index(if_index))

    def connect_servo_canopen(
        self,
        can_device: CanDevice,
        dict_path: str,
        node_id: int,
        baudrate: CanBaudrate = CanBaudrate.Baudrate_1M,
        channel: int = 0,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> tuple[CanopenNetwork, CanopenServo]:
        """Connect to target servo by CANOpen.

        Args:
            can_device : CANOpen device type.
            dict_path : servo dictionary path.
            node_id : node id. It's posible scan node ids with
                :func:`scan_servos_canopen`.
            baudrate : communication baudrate. 1 Mbit/s by default.
            channel : CANopen device channel. ``0`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Returns:
            The CANOpen network and the connected servo.

        Raises:
            FileNotFoundError: If either of the dict files doesn't exist.
            ingenialink.exceptions.ILError: If CANOpen device type, node id or channel is incorrect.
        """
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"Dict file {dict_path} does not exist!")

        net_key = f"{can_device}_{channel}_{baudrate}"
        if net_key not in self.mc.net:
            self.mc.net[net_key] = CanopenNetwork(can_device, channel, baudrate)
        net = cast("CanopenNetwork", self.mc.net[net_key])

        servo = net.connect_to_slave(
            node_id,
            dict_path,
            servo_status_listener,
            net_status_listener,
            disconnect_callback=self.__disconnect_callback,
        )
        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = net_key
        return net, servo

    def connect_servo_ethercat(
        self,
        interface_name: str,
        slave_id: int,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
        gil_release_config: GilReleaseConfig = GilReleaseConfig(),
    ) -> tuple[EthercatNetwork, "EthercatServo"]:
        r"""Connect to an EtherCAT slave - CoE.

        Args:
            interface_name : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            slave_id: EtherCAT slave ID.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.
            gil_release_config: GIL release configuration.

        Returns:
            The EtherCAT network and the connected servo.

        Raises:
            FileNotFoundError: If the dict file doesn't exist.
        """
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"Dict file {dict_path} does not exist!")
        if interface_name not in self.mc.net:
            self.mc.net[interface_name] = EthercatNetwork(
                interface_name, gil_release_config=gil_release_config
            )
        net = cast("EthercatNetwork", self.mc.net[interface_name])
        try:
            servo = net.connect_to_slave(
                slave_id,
                dict_path,
                servo_status_listener=servo_status_listener,
                net_status_listener=net_status_listener,
                disconnect_callback=self.__disconnect_callback,
            )
        except ILError as e:
            if len(net.servos) == 0:
                del self.mc.net[interface_name]
            raise e
        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = interface_name
        return net, servo

    def connect_servo_ethercat_interface_index(
        self,
        if_index: int,
        slave_id: int,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
        gil_release_config: GilReleaseConfig = GilReleaseConfig(),
    ) -> tuple[EthercatNetwork, "EthercatServo"]:
        """Connect to an EtherCAT slave - CoE.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
            slave_id: EtherCAT slave ID.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.
            gil_release_config: GIL release configuration.

        Returns:
            The EtherCAT network and the connected servo.

        Raises:
            IndexError: If interface index is out of range.
        """
        return self.connect_servo_ethercat(
            self.get_ifname_by_index(if_index),
            slave_id,
            dict_path,
            alias,
            servo_status_listener,
            net_status_listener,
            gil_release_config=gil_release_config,
        )

    def connect_servo_ethercat_interface_ip(
        self,
        interface_ip: str,
        slave_id: int,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
        gil_release_config: GilReleaseConfig = GilReleaseConfig(),
    ) -> tuple[EthercatNetwork, "EthercatServo"]:
        """Connect to an EtherCAT slave - CoE.

        Args:
            interface_ip : IP of the interface to be connected to.
            slave_id: EtherCAT slave ID.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.
            gil_release_config: GIL release configuration.

        Returns:
            The EtherCAT network and the connected servo.
        """
        return self.connect_servo_ethercat(
            self.get_ifname_from_interface_ip(interface_ip),
            slave_id,
            dict_path,
            alias,
            servo_status_listener,
            net_status_listener,
            gil_release_config=gil_release_config,
        )

    @staticmethod
    def scan_servos_ethercat_with_info(
        interface_name: str, gil_release_config: GilReleaseConfig = GilReleaseConfig()
    ) -> OrderedDict[int, SlaveInfo]:
        r"""Scan a network adapter.

         Get all connected EtherCAT slaves including slave information.

        Args:
            interface_name : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            gil_release_config: GIL release configuration.

        Returns:
            Dictionary of nodes available in the network and slave information.

        Raises:
            TypeError: If some parameter has a wrong type.
        """
        net = EthercatNetwork(interface_name, gil_release_config=gil_release_config)
        slaves_info = net.scan_slaves_info()
        return slaves_info

    def scan_servos_ethercat(
        self, interface_name: str, gil_release_config: GilReleaseConfig = GilReleaseConfig()
    ) -> list[int]:
        r"""Scan a network adapter to get all connected EtherCAT slaves.

        Args:
            interface_name : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            gil_release_config: GIL release configuration.

        Returns:
            List of EtherCAT slaves available in the network.

        Raises:
            TypeError: If some parameter has a wrong type.
        """
        net = EthercatNetwork(interface_name, gil_release_config=gil_release_config)
        slaves = net.scan_slaves()
        return slaves

    def scan_servos_ethercat_interface_ip(self, interface_ip: str) -> list[int]:
        """Scan a network adapter to get all connected EtherCAT slaves.

        Args:
            interface_ip : IP of the interface to be connected to.

        Returns:
            List of EtherCAT slaves available in the network.

        """
        return self.scan_servos_ethercat(self.get_ifname_from_interface_ip(interface_ip))

    def scan_servos_ethercat_interface_index(self, if_index: int) -> list[int]:
        """Scan a network adapter to get all connected EtherCAT slaves.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.

        Returns:
            List of EtherCAT slaves available in the network.

        Raises:
            IndexError: If interface index is out of range.

        """
        return self.scan_servos_ethercat(self.get_ifname_by_index(if_index))

    def scan_servos_canopen_with_info(
        self,
        can_device: CanDevice,
        baudrate: CanBaudrate = CanBaudrate.Baudrate_1M,
        channel: int = 0,
    ) -> OrderedDict[int, SlaveInfo]:
        """Scan CANOpen device network to get all nodes including slave information.

        Args:
            can_device : CANOpen device type.
            baudrate : communication baudrate. 1 Mbit/s by default.
            channel : CANOpen device channel. ``0`` by default.

        Returns:
            Dictionary of nodes available in the network and slave information.

        Raise:
            TypeError: If some parameter has a wrong type.
        """
        net_key = f"{can_device}_{channel}_{baudrate}"
        if net_key not in self.mc.net:
            self.mc.net[net_key] = CanopenNetwork(can_device, channel, baudrate)
        net = self.mc.net[net_key]

        if net is None:
            self.logger.warning(
                "Could not find any nodes in the network.Device: %s, channel: %s and baudrate: %s.",
                can_device,
                channel,
                baudrate,
            )
            return []
        slaves_info = net.scan_slaves_info()
        return slaves_info

    def scan_servos_canopen(
        self,
        can_device: CanDevice,
        baudrate: CanBaudrate = CanBaudrate.Baudrate_1M,
        channel: int = 0,
    ) -> list[int]:
        """Scan CANOpen device network to get all nodes.

        Args:
            can_device : CANOpen device type.
            baudrate : communication baudrate. 1 Mbit/s by default.
            channel : CANOpen device channel. ``0`` by default.

        Returns:
            List of node ids available in the network.

        Raises:
            TypeError: If some parameter has a wrong type.
        """
        net_key = f"{can_device}_{channel}_{baudrate}"
        if net_key not in self.mc.net:
            self.mc.net[net_key] = CanopenNetwork(can_device, channel, baudrate)
        net = self.mc.net[net_key]

        if net is None:
            self.logger.warning(
                "Could not find any nodes in the network.Device: %s, channel: %s and baudrate: %s.",
                can_device,
                channel,
                baudrate,
            )
            return []
        slaves = net.scan_slaves()
        return slaves

    @staticmethod
    def scan_servos_ethernet(
        subnet: str,
    ) -> list[str]:
        """Scan Ethernet device network to get all nodes.

        Args:
            subnet: The subnet in CIDR notation. For instance ``192.168.1.1/24``.

        Returns:
            List of drive IPs available in the network.

        """
        net = EthernetNetwork(subnet)
        slaves = net.scan_slaves()
        return slaves

    @staticmethod
    def scan_servos_ethernet_with_info(
        subnet: str,
    ) -> OrderedDict[str, SlaveInfo]:
        """Scan Ethernet device network to get all nodes including drive information.

        Args:
            subnet: The subnet in CIDR notation. For instance ``192.168.1.1/24``.

        Returns:
            Dictionary of nodes available in the network and drive information.

        """
        net = EthernetNetwork(subnet)
        slaves_info = net.scan_slaves_info()
        return slaves_info

    def disconnect(self, servo: str = DEFAULT_SERVO) -> None:
        """Disconnect servo.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        network = self.mc._get_network(servo)
        # This will call `__disconnect_callback` to complete the disconnection
        network.disconnect_from_slave(drive)

    def get_servo_state(self, servo: str = DEFAULT_SERVO) -> NetState:
        """Get the network state of a servo (connected/disconnected).

        The net_status_listener should be enabled for the servo in order
        for the state to be updated.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        Returns:
            The servo's state.

        """
        drive = self.mc._get_drive(servo)
        network = self.mc._get_network(servo)
        return network.get_servo_state(drive.target)

    def get_register(
        self, register: str, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Union[int, float, str]:
        """Return the value of a target register.

        Args:
            register : register UID.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Current register value.

        Raises:
            ingenialink.exceptions.ILAccessError: If the register access is write-only.
            IMRegisterNotExistError: If the register doesn't exist.
            TypeError: If some parameter has a wrong type.

        """
        drive = self.mc._get_drive(servo)
        register_dtype = self.mc.info.register_type(register, axis, servo=servo)
        value = drive.read(register, subnode=axis)
        if register_dtype.value <= RegDtype.S64.value and isinstance(value, int):
            return int(value)
        if not isinstance(value, (int, float, str)):
            raise TypeError("Register value is not a correct type of value.")
        return value

    def set_register(
        self,
        register: str,
        value: Union[int, float, str],
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set a value of a target register.

        Args:
            register : register UID.
            value : new value for the register.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If the value is of the wrong type.
            IMRegisterNotExistError: If the register doesn't exist.
            IMRegisterWrongAccessError: If the register access is read-only.

        """
        drive = self.mc._get_drive(servo)
        register_dtype_value = self.mc.info.register_type(register, axis, servo=servo)
        register_access_type = self.mc.info.register_info(register, axis, servo=servo).access
        signed_int = [RegDtype.S8, RegDtype.S16, RegDtype.S32, RegDtype.S64]
        unsigned_int = [RegDtype.U8, RegDtype.U16, RegDtype.U32, RegDtype.U64]
        if register_dtype_value == RegDtype.FLOAT and not isinstance(value, (int, float)):
            raise TypeError("Value must be a float")
        if register_dtype_value == RegDtype.STR and not isinstance(value, str):
            raise TypeError("Value must be a string")
        if register_dtype_value in signed_int and not isinstance(value, int):
            raise TypeError("Value must be an int")
        if register_dtype_value in unsigned_int and (not isinstance(value, int) or value < 0):
            raise TypeError("Value must be an unsigned int")
        if register_access_type == RegAccess.RO:
            raise IMRegisterWrongAccessError(
                f"Register: {register} cannot write to a read-only register"
            )
        drive.write(register, value, subnode=axis)

    def subscribe_net_status(
        self, callback: Callable[[NetDevEvt], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Add a callback to net status change event.

        Args:
            callback : when net status changes callback is called.
            servo : servo alias to reference it. ``default`` by default.

        """
        network = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        network.subscribe_to_status(drive.target, callback)

    def unsubscribe_net_status(
        self, callback: Callable[[NetDevEvt], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Remove net status change event callback.

        Args:
            callback : callback to remove.
            servo : servo alias to reference it. ``default`` by default.

        """
        network = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        network.unsubscribe_from_status(drive.target, callback)

    def subscribe_servo_status(
        self, callback: Callable[[ServoState, int], Any], servo: str = DEFAULT_SERVO
    ) -> None:
        """Add a callback to servo status change event.

        Args:
            callback : when servo status changes callback is called.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        drive.subscribe_to_status(callback)

    def unsubscribe_servo_status(
        self, callback: Callable[[ServoState, int], Any], servo: str = DEFAULT_SERVO
    ) -> None:
        """Remove servo status change event callback.

        Args:
            callback : callback to remove.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        drive.unsubscribe_from_status(callback)

    def subscribe_register_update(
        self,
        callback: Callable[[str, Servo, Register, Union[int, float, str, bytes]], None],
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Subscribe to register updates.

        The callback will be called when a read/write operation occurs.

        Args:
            callback: Callable that takes a servo alias, Servo and Register instances
            and the register value as arguments.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)

        if drive not in self.register_update_observers:
            # Servo that has not been subscribed yet
            self.register_update_observers[drive] = []
            # Subscribe to ingenialink
            drive.register_update_subscribe(self._il_register_subscribe_callback)

        self.register_update_observers[drive].append(
            IMRegisterUpdateObserver(callback, alias=servo)
        )

    def unsubscribe_register_update(
        self,
        callback: Callable[[str, Servo, Register, Union[int, float, str, bytes]], None],
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Unsubscribe from register updates.

        Args:
            callback: Subscribed callback.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        for observer in self.register_update_observers[drive]:
            if observer.im_callback == callback:
                self.register_update_observers[drive].remove(observer)
                break

        if len(self.register_update_observers[drive]) == 0:
            del self.register_update_observers[drive]
            # No observers, unsubscribe from ingenialink
            drive.register_update_unsubscribe(self._il_register_subscribe_callback)

    def _il_register_subscribe_callback(
        self, servo_instance: Servo, register: Register, value: Union[int, float, str, bytes]
    ) -> None:
        """This method will be the one subscribed to ingenialink.

        When called, the servo alias will be added to the received information.

        Args:
            servo_instance: The servo instance.
            register: The register object.
            value: The updated register value.

        """
        for observer in self.register_update_observers[servo_instance]:
            observer.im_callback(observer.alias, servo_instance, register, value)

    def subscribe_emergency_message(
        self,
        callback: Callable[[str, EmergencyMessage], None],
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Subscribe to emergency messages.

        Only available for CANopen and EtherCAT CoE protocols.

        Args:
            callback :  Callable that takes a servo alias and an EmergencyMessage
            instance as arguments.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)

        if drive not in self.emergency_messages_observers:
            # Servo that has not been subscribed yet
            self.emergency_messages_observers[drive] = []
            # Subscribe to ingenialink
            drive.emcy_subscribe(self._il_emergency_message_callback)

        self.emergency_messages_observers[drive].append(
            IMEmergencyMessageObserver(callback, alias=servo)
        )

    def unsubscribe_emergency_message(
        self,
        callback: Callable[[str, EmergencyMessage], None],
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Unsubscribe from emergency messages.

        Only available for CANopen and EtherCAT CoE protocols.

        Args:
            callback : Subscribed callback.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        for observer in self.emergency_messages_observers[drive]:
            if observer.im_callback == callback:
                self.emergency_messages_observers[drive].remove(observer)
                break

        if len(self.emergency_messages_observers[drive]) == 0:
            del self.emergency_messages_observers[drive]
            # No observers, unsubscribe from ingenialink
            drive.emcy_unsubscribe(self._il_emergency_message_callback)

    def _il_emergency_message_callback(self, emergency_message: EmergencyMessage) -> None:
        """This method will be the one subscribed to ingenialink.

        When called, the servo alias will be added to the received information.

        Args:
            emergency_message: The EmergencyMessage  instance.

        """
        for observer in self.emergency_messages_observers[emergency_message.servo]:
            observer.im_callback(observer.alias, emergency_message)

    def load_firmware_canopen(
        self,
        fw_file: str,
        servo: str = DEFAULT_SERVO,
        status_callback: Optional[Callable[[str], None]] = None,
        progress_callback: Optional[Callable[[int], None]] = None,
        error_enabled_callback: Optional[Callable[[bool], None]] = None,
    ) -> None:
        """Load firmware via CANopen.

        Args:
            fw_file : Firmware file path.
            servo : servo alias to reference it. ``default`` by default.
            status_callback : callback with status.
            progress_callback : callback with progress.
            error_enabled_callback : callback with errors enabled.

        Raises:
            ValueError: If servo is not connected via CANopen.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        if not isinstance(net, CanopenNetwork):
            raise ValueError("Target servo is not connected via CANopen")
        if not isinstance(drive, CanopenServo):
            raise ValueError("Target servo is not a Canopen servo")
        if status_callback is None:
            status_callback = partial(self.logger.info, "Load firmware status: %s")
        if progress_callback is None:
            progress_callback = partial(self.logger.info, "Load firmware progress: %s")
        if fw_file.endswith(self.ENSEMBLE_FIRMWARE_EXTENSION):
            self.__load_ensemble_fw_canopen(
                net,
                fw_file,
                drive,
                status_callback,
                progress_callback,
                error_enabled_callback,
            )
        else:
            net.load_firmware(
                int(drive.target),
                fw_file,
                status_callback,
                progress_callback,
                error_enabled_callback,
            )

    @staticmethod
    def __get_boot_in_app(fw_file: str) -> bool:
        """Return true if the bootloader is included in the application (file extension is .sfu).

        Args:
            fw_file: Path to the FW file.

        Returns:
            True if the bootloader is included in the application.

        Raises:
            ValueError: If the firmware file has the wrong extension.
        """
        if not fw_file.endswith((FILE_EXT_SFU, FILE_EXT_LFU)):
            raise ValueError(
                f"Firmware file should have extension {FILE_EXT_SFU} or {FILE_EXT_LFU}."
            )
        return fw_file.endswith(FILE_EXT_SFU)

    def load_firmware_ecat(
        self,
        ifname: str,
        fw_file: str,
        slave: int = 1,
        boot_in_app: Optional[bool] = None,
        password: Optional[int] = None,
        gil_release_config: GilReleaseConfig = GilReleaseConfig(),
    ) -> None:
        r"""Load firmware via ECAT.

        Args:
            ifname : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            fw_file : Firmware file path.
            slave : slave index. ``1`` by default.
            boot_in_app: true if the bootloader is included in the application, false otherwise.
                If None, the file extension is used to define it.
            password: Password to load the firmware file. If ``None`` the default password will be
                used.
            gil_release_config: GIL release config.

        Raises:
            FileNotFoundError: If the firmware file cannot be found.
            ValueError: If the firmware file has the wrong extension.
            ingenialink.exceptions.ILFirmwareLoadError: If no slave is detected.
            ingenialink.exceptions.ILFirmwareLoadError: If the FoE write operation fails.

        """
        net = EthercatNetwork(ifname, gil_release_config=gil_release_config)
        if fw_file.endswith(self.ENSEMBLE_FIRMWARE_EXTENSION):
            self.__load_ensemble_fw_ecat(net, fw_file, slave, boot_in_app, password)
        else:
            boot_in_app = self.__get_boot_in_app(fw_file) if boot_in_app is None else boot_in_app
            net.load_firmware(fw_file, boot_in_app, slave, password)

    def load_firmware_ecat_interface_index(
        self,
        if_index: int,
        fw_file: str,
        slave: int = 1,
        boot_in_app: Optional[bool] = None,
        password: Optional[int] = None,
        gil_release_config: GilReleaseConfig = GilReleaseConfig(),
    ) -> None:
        """Load firmware via ECAT.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
            fw_file : Firmware file path.
            slave : slave index. ``1`` by default.
            boot_in_app: true if the bootloader is included in the application, false otherwise.
                If ``None``, the file extension is used to define it.
            password: Password to load the firmware file. If ``None`` the default password will be
                used.
            gil_release_config: GIL release config.

        Raises:
            IndexError: If interface index is out of range.
            FileNotFoundError: If the firmware file cannot be found.
            ingenialink.exceptions.ILFirmwareLoadError: If no slave is detected.
            ingenialink.exceptions.ILFirmwareLoadError: If the FoE write operation is not successful
            NotImplementedError: If FoE is not implemented for the current OS and architecture

        """
        self.load_firmware_ecat(
            self.get_ifname_by_index(if_index),
            fw_file,
            slave,
            boot_in_app,
            password,
            gil_release_config=gil_release_config,
        )

    def load_firmware_ethernet(
        self, ip: str, fw_file: str, ftp_user: Optional[str] = None, ftp_pwd: Optional[str] = None
    ) -> None:
        """Load firmware via Ethernet. Boot mode is needed to load firmware.

        .. warning::
            After functions ends, the servo will take a moment to load firmware.
            During the process, the servo will be not operative.

        Args:
            ip : servo IP.
            fw_file : Firmware file path.
            ftp_user : FTP user to connect with.
            ftp_pwd : FTP password for the given user.

        """
        net = EthernetNetwork()
        ftp_user = ftp_user or "Ingenia"
        ftp_pwd = ftp_pwd or "Ingenia"
        net.load_firmware(fw_file, ip, ftp_user, ftp_pwd)

    @staticmethod
    def __ftp_ping(ip: str) -> bool:
        response = ping(ip, timeout=1)
        return isinstance(response, float)

    def boot_mode_and_load_firmware_ethernet(
        self,
        fw_file: str,
        servo: str = DEFAULT_SERVO,
        ftp_user: Optional[str] = None,
        ftp_pwd: Optional[str] = None,
    ) -> None:
        """Set servo to boot mode and load firmware. Servo is disconnected.

        .. warning::
            After functions ends, the servo will take a moment to load firmware.
            During the process, the servo will be not operative.

        Args:
            fw_file : Firmware file path.
            servo : servo alias to reference it. ``default`` by default.
            ftp_user : FTP user to connect with.
            ftp_pwd : FTP password for the given user.

        Raises:
            ValueError: If servo is not connected via Ethernet.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        ip = str(drive.target)
        if not isinstance(net, EthernetNetwork):
            raise ValueError("Target servo is not connected via Ethernet")
        self.boot_mode(servo)
        timeout = 5
        init_time = time.time()
        ftp_ready = False
        while not ftp_ready and time.time() - init_time < timeout:
            ftp_ready = self.__ftp_ping(ip)
            time.sleep(1)
        self.load_firmware_ethernet(ip, fw_file, ftp_user, ftp_pwd)

    def boot_mode(self, servo: str = DEFAULT_SERVO) -> None:
        """Set servo to boot mode. Servo is disconnected.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        net.stop_status_listener()
        drive.stop_status_listener()
        with contextlib.suppress(ILError):
            self.mc.communication.set_register(
                self.FORCE_SYSTEM_BOOT_COCO_REGISTER,
                self.PASSWORD_FORCE_BOOT_COCO,
                servo=servo,
                axis=0,
            )
        self.disconnect(servo)

    def load_firmware_moco(
        self,
        fw_file: str,
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Load firmware to the Motion Core.

        .. warning::
            After functions ends, the servo will take a moment to load firmware.
            During the process, the servo will be not operative.

        Args:
            fw_file : Firmware file path.
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If servo is not connected via Ethernet.

        """
        default_node = 10
        default_subnode = 1
        default_port = 1061
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        ip = str(drive.target)
        if not isinstance(net, EthernetNetwork):
            raise ValueError("Target servo is not connected via Ethernet")
        net.load_firmware_moco(default_node, default_subnode, ip, default_port, fw_file)

    def boot_mode_moco(self, servo: str = DEFAULT_SERVO) -> None:
        """Set the Motion Core to boot mode.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        net.stop_status_listener()
        drive.stop_status_listener()
        try:
            self.mc.communication.set_register(
                self.FORCE_SYSTEM_BOOT_MOCO_REGISTER,
                self.PASSWORD_FORCE_BOOT_MOCO,
                servo=servo,
            )
        except ILError as e:
            self.logger.debug(e)
        try:
            self.mc.communication.set_register(
                self.SYSTEM_RESET_MOCO_REGISTER,
                self.PASSWORD_FORCE_BOOT_MOCO,
                servo=servo,
            )
        except ILError as e:
            self.logger.debug(e)

    def __load_ensemble_fw_canopen(
        self,
        net: CanopenNetwork,
        fw_file: str,
        slave: CanopenServo,
        status_callback: Optional[Callable[[str], None]] = None,
        progress_callback: Optional[Callable[[int], None]] = None,
        error_enabled_callback: Optional[Callable[[bool], None]] = None,
    ) -> None:
        """Load FW to an ensemble of servos through Canopen (in parallel).

        The FW files that are contained in the fw_file will be loaded by selecting any slave which
        is part of the ensemble.

        Args:
            net: Canopen network.
            fw_file: Path to the ensemble FW file.
            slave: Servo object.
            status_callback : callback with status.
            progress_callback : callback with progress.
            error_enabled_callback : callback with errors enabled.

        Raises:
            IMFirmwareLoadError: If the load FW process of any slave failed.
        """
        with tempfile.TemporaryDirectory() as ensemble_temp_dir:
            mapping = self.__unzip_ensemble_fw_file(fw_file, ensemble_temp_dir)
            scanned_slaves = net.scan_slaves_info()
            first_slave_in_ensemble = self.__check_ensemble(
                scanned_slaves, int(slave.target), mapping
            )
            dictionary_path = slave.dictionary.path
            connected_drives = {int(slave.target): slave for slave in net.servos}
            for slave_id_offset in mapping:
                slave_id = first_slave_in_ensemble + slave_id_offset
                if slave_id not in connected_drives:
                    net.connect_to_slave(
                        slave_id, dictionary_path, disconnect_callback=self.__disconnect_callback
                    )

            load_fw_slave_id: Optional[int] = None
            try:
                for slave_id_offset, fw_file_prod_code in mapping.items():
                    load_fw_slave_id = first_slave_in_ensemble + slave_id_offset
                    net.load_firmware(
                        load_fw_slave_id,
                        fw_file_prod_code[0],
                        status_callback,
                        progress_callback,
                        error_enabled_callback,
                    )
            except ILError as e:
                if load_fw_slave_id is None:
                    raise IMFirmwareLoadError(
                        f"{FIRMWARE_FILE_FAIL_MSG}. No slave detected in the ensemble."
                    )
                raise IMFirmwareLoadError(
                    f"{FIRMWARE_FILE_FAIL_MSG} on node {load_fw_slave_id}. Exception: {e}"
                )

    def __load_ensemble_fw_ecat(
        self,
        net: EthercatNetwork,
        fw_file: str,
        slave: int,
        boot_in_app: Optional[bool],
        password: Optional[int],
    ) -> None:
        """Load FW to an ensemble of servos through Ethercat.

        The FW files that are contained in the fw_file will be loaded by selecting any slave which
        is part of the ensemble.

        Args:
            net: Ethercat network.
            fw_file: Path to the ensemble FW file.
            slave: Slave ID (any slave in the ensemble)
            boot_in_app: true if the bootloader is included in the application, false otherwise.
                If None, the file extension is used to define it.
            password: Password to load the firmware file. If ``None`` the default password will be
                used.

        Raises:
            IMFirmwareLoadError: If the load FW process of any slave failed.
        """
        with tempfile.TemporaryDirectory() as ensemble_temp_dir:
            mapping = self.__unzip_ensemble_fw_file(fw_file, ensemble_temp_dir)
            scanned_slaves = net.scan_slaves_info()
            if len(scanned_slaves) == 0:
                raise IMFirmwareLoadError(f"{FIRMWARE_FILE_FAIL_MSG}. No ECAT slave detected.")
            first_slave_in_ensemble = self.__check_ensemble(scanned_slaves, slave, mapping)
            try:
                for slave_id_offset, fw_file_prod_code in mapping.items():
                    boot_in_app_drive = (
                        self.__get_boot_in_app(fw_file_prod_code[0])
                        if boot_in_app is None
                        else boot_in_app
                    )
                    net.load_firmware(
                        fw_file_prod_code[0],
                        boot_in_app_drive,
                        first_slave_in_ensemble + slave_id_offset,
                        password,
                    )
            except ILError as e:
                raise e

    def __check_ensemble(
        self,
        scanned_slaves: OrderedDict[int, SlaveInfo],
        slave_id: int,
        mapping: dict[int, tuple[str, int, int]],
    ) -> int:
        """Check the ensemble.

        Check that a slave is part of the ensemble described in the mapping argument and the
        ensemble is complete (all drives described in the mapping are contained in the list of
        scanned slaves).

        It returns the ID of the first drive in the ensemble.

        Args:
            scanned_slaves: Scanned slaves with the respective information.
            slave_id: Slave ID to be checked.
            mapping: Mapping of the ensemble.

        Raises:
            IMFirmwareLoadError: If the slave ID is not in the scanned slaves list.
            IMFirmwareLoadError: If the ensemble described in the mapping can not be
            found in the list of the scanned slaves.

        Returns:
            The ID of the first drive in the ensemble.
        """
        if slave_id not in scanned_slaves:
            raise IMFirmwareLoadError(
                f"{FIRMWARE_FILE_FAIL_MSG}. The slave {slave_id} is not detected."
            )
        slave_id_offset = self.__check_slave_in_ensemble(scanned_slaves[slave_id], mapping)
        first_slave = slave_id - slave_id_offset
        for map_slave_id_offset in mapping:
            map_slave_id = first_slave + map_slave_id_offset
            if map_slave_id not in scanned_slaves:
                raise IMFirmwareLoadError(
                    f"{FIRMWARE_FILE_FAIL_MSG}. The slave {map_slave_id - 1} is not detected."
                )
            map_slave_info = scanned_slaves[map_slave_id]
            if (
                map_slave_info.product_code != mapping[map_slave_id_offset][1]
                or map_slave_info.revision_number is None
                or (map_slave_info.revision_number & self.ENSEMBLE_SLAVE_REV_NUM_MASK)
                != (mapping[map_slave_id_offset][2] & self.ENSEMBLE_SLAVE_REV_NUM_MASK)
            ):
                raise IMFirmwareLoadError(
                    f"{FIRMWARE_FILE_FAIL_MSG}. The slave {map_slave_id} "
                    f"has wrong product code or revision number."
                )
        return first_slave

    def __check_slave_in_ensemble(
        self, slave_info: SlaveInfo, mapping: dict[int, tuple[str, int, int]]
    ) -> int:
        """Check that a slave is part of the ensemble described in the mapping argument.

        Returns the ID offset (relative position in the ensemble) of the selected slave.

        Args:
            slave_info: Info of the slave to be checked.
            mapping: Mapping of the ensemble.

        Raises:
            IMFirmwareLoadError: If the slave is not part of the ensemble.

        Returns:
            The ID offset (relative position in the ensemble) of the selected slave.
        """
        for slave_id_offset in mapping:
            _, mapping_product_code, mapping_revision_number = mapping[slave_id_offset]
            scanned_product_code = slave_info.product_code
            scanned_revision_number = slave_info.revision_number
            if scanned_product_code is None or scanned_revision_number is None:
                continue
            if mapping_product_code == scanned_product_code and (
                mapping_revision_number & self.ENSEMBLE_SLAVE_REV_NUM_MASK
            ) == (scanned_revision_number & self.ENSEMBLE_SLAVE_REV_NUM_MASK):
                return slave_id_offset
        raise IMFirmwareLoadError(
            f"{FIRMWARE_FILE_FAIL_MSG}. The selected drive is not part of the ensemble."
        )

    def __unzip_ensemble_fw_file(
        self, fw_file: str, unzip_path: str
    ) -> dict[int, tuple[str, int, int]]:
        """Unzip the ensemble FW file and return the mapping.

        Args:
            fw_file: Ensemble FW file to be unzipped.
            unzip_path: Path where to unzip the ensemble

        Returns:
            Mapping described in the ensemble FW file.
                Dict{slave_id_offset: (fw_file, product_code, revision_number)}
        """
        with zipfile.ZipFile(fw_file, "r") as zip_ref:
            zip_ref.extractall(unzip_path)
        with open(path.join(unzip_path, "mapping.json")) as f:
            mapping_info = json.load(f)
        mapping = {}
        for drive in mapping_info["drives"]:
            slave_id_offset = drive["slave_id_offset"]
            fw_file = path.join(path.abspath(unzip_path), drive["fw_file"])
            product_code = drive["product_code"]
            revision_number = drive["revision_number"]
            mapping[slave_id_offset] = fw_file, product_code, revision_number

        return mapping



from typing import TYPE_CHECKING, Optional

import ingenialogger
from ingenialink.pdo import PDOMap, RPDOMap, RPDOMapItem, TPDOMap, TPDOMapItem

from ingeniamotion.fsoe_master.frame import (
    MASTER_FRAME_ELEMENTS,
    SLAVE_FRAME_ELEMENTS,
    FSoEFrame,
)
from ingeniamotion.fsoe_master.fsoe import (
    FSoEDictionary,
    FSoEDictionaryItem,
    FSoEDictionaryItemInput,
    FSoEDictionaryItemInputOutput,
    FSoEDictionaryItemOutput,
    FSoEDictionaryMap,
)
from ingeniamotion.fsoe_master.maps_validator import (
    FSoEDictionaryMapValidator,
    FSoEFrameRules,
    FSoEFrameRuleValidatorOutput,
)
from ingeniamotion.fsoe_master.safety_functions import STOFunction

if TYPE_CHECKING:
    from ingenialink.dictionary import Dictionary

    from ingeniamotion.fsoe_master import FSoEMasterHandler
    from ingeniamotion.fsoe_master.safety_functions import SafetyFunction

__all__ = ["ProcessImage"]


class ProcessImage:
    """Helper class to configure the Safety Process Image."""

    __SLOT_WIDTH = 16
    """Number of bits in a data slot of the Safety PDU."""

    def __init__(
        self,
        outputs: "FSoEDictionaryMap",
        inputs: "FSoEDictionaryMap",
    ) -> None:
        self.outputs = outputs
        self.inputs = inputs
        self.__validator: FSoEDictionaryMapValidator = FSoEDictionaryMapValidator()
        self.logger = ingenialogger.get_logger(__name__)

    @classmethod
    def empty(cls, dictionary: "FSoEDictionary") -> "ProcessImage":
        """Create an empty ProcessImage instance with the given dictionary.

        Returns:
            ProcessImage instance with empty outputs and inputs maps.
        """
        return cls(
            outputs=FSoEDictionaryMap(
                dictionary,
                item_types_accepted={FSoEDictionaryItemOutput, FSoEDictionaryItemInputOutput},
            ),
            inputs=FSoEDictionaryMap(
                dictionary,
                item_types_accepted={FSoEDictionaryItemInput, FSoEDictionaryItemInputOutput},
            ),
        )

    @classmethod
    def default(cls, dictionary: "FSoEDictionary") -> "ProcessImage":
        """Create a default ProcessImage instance with the default dictionary.

        Returns:
            ProcessImage instance with the minimum required items.
        """
        process_image = cls.empty(dictionary)
        process_image.outputs.add(dictionary.name_map[STOFunction.COMMAND_UID])
        process_image.inputs.add(dictionary.name_map[STOFunction.COMMAND_UID])
        return process_image

    def copy(self) -> "ProcessImage":
        """Create a copy of the ProcessImage instance.

        Returns:
            A new ProcessImage instance with copies of the outputs and inputs maps.
        """
        return ProcessImage(
            outputs=self.outputs.copy(),
            inputs=self.inputs.copy(),
        )

    @property
    def editable(self) -> bool:
        """Indicates if the Process image can be edited."""
        return not (self.outputs.locked or self.inputs.locked)

    def __validate_dictionary_map(
        self,
        dictionary_map: "FSoEDictionaryMap",
        rules: Optional[list[FSoEFrameRules]],
        raise_exceptions: bool,
    ) -> FSoEFrameRuleValidatorOutput:
        """Validate the given dictionary map against the specified rules.

        Args:
            dictionary_map: The dictionary map to validate.
            rules: Optional list of specific rules to validate. If None, all rules are validated.
            raise_exceptions: If True, raises an exception if any rule is invalid.
                If False, returns the validation output with exceptions. Defaults to False.

        Returns:
            The output of the validation containing the rules and exceptions.
        """
        return self.__validator.validate_dictionary_map_fsoe_frame_rules(
            dictionary_map, rules, raise_exceptions
        )

    def are_inputs_valid(
        self, rules: Optional[list[FSoEFrameRules]] = None, raise_exceptions: bool = False
    ) -> FSoEFrameRuleValidatorOutput:
        """Check if the inputs map is valid.

        Args:
            rules: list of specific rules to validate. If None, all rules are validated.
            raise_exceptions: If True, raises an exception if any rule is invalid.
                If False, returns the validation output with exceptions. Defaults to False.

        Returns:
            The output of the validation containing the rules and exceptions.
        """
        return self.__validate_dictionary_map(self.inputs, rules, raise_exceptions)

    def are_outputs_valid(
        self, rules: Optional[list[FSoEFrameRules]] = None, raise_exceptions: bool = False
    ) -> FSoEFrameRuleValidatorOutput:
        """Check if the outputs map is valid.

        Args:
            rules: list of specific rules to validate. If None, all rules are validated.
            raise_exceptions: If True, raises an exception if any rule is invalid.
                If False, returns the validation output with exceptions. Defaults to False.

        Returns:
            The output of the validation containing the rules and exceptions.
        """
        return self.__validate_dictionary_map(self.outputs, rules, raise_exceptions)

    def validate(self) -> bool:
        """Check if the map is valid.

        Args:
            rules: list of specific rules to validate. If None, all rules are validated.
            raise_exceptions: If True, raises an exception if any rule is invalid.
                If False, returns the validation output with exceptions. Defaults to False.

        Returns:
            True if the maps are valid.
        """
        self.are_inputs_valid(rules=None, raise_exceptions=True)
        self.are_outputs_valid(rules=None, raise_exceptions=True)
        return True

    def insert_in_best_position(self, element: "FSoEDictionaryItem") -> None:
        """Insert I/O in any best position of the maps.

        Finds unused space in the map and insert them there.
        If not unused space is found, it appends the item to the end of the map.

        Args:
            element: element to add

        Raises:
            TypeError: if the element is not of a valid type.
        """
        align_to = 1
        if element.data_type.bits > 8:
            align_to = 8
        # Objects larger than 16-bit must be word-aligned
        if element.data_type.bits > 16:
            align_to = 16
        if isinstance(element, FSoEDictionaryItemOutput):
            self.outputs.insert_in_best_position(element, align_to)
        elif isinstance(element, FSoEDictionaryItemInput):
            self.inputs.insert_in_best_position(element, align_to)
        elif isinstance(element, FSoEDictionaryItemInputOutput):
            self.inputs.insert_in_best_position(element, align_to)
            self.outputs.insert_in_best_position(element, align_to)
        else:
            raise TypeError(f"Unknown IO Element type: {type(element)}")

    def insert_safety_function(self, safety_function: "SafetyFunction") -> None:
        """Insert all elements of the safety function on the maps.

        Args:
            safety_function: Safety function
        """
        for io in safety_function.ios.values():
            self.insert_in_best_position(io)

    def insert_safety_functions_by_type(
        self, handler: "FSoEMasterHandler", function_type: type["SafetyFunction"]
    ) -> "SafetyFunction":
        """Insert the first non-mapped instance of the given safety function type on the maps.

        Args:
            handler: FSoE handler.
            function_type: Safety function type.

        Returns:
            The safety function that was inserted.

        Raises:
            ValueError: if no safety functions of the given type are found.
            ValueError: if all safety functions of the given type are already mapped.

        """
        sf_available: list[SafetyFunction] = handler.safety_functions_by_type().get(
            function_type, []
        )
        if not sf_available:
            raise ValueError(f"No safety functions of type {function_type} found")
        for function in sf_available:
            if not self.is_safety_function_mapped(function):
                self.unmap_safety_function(function)
                self.insert_safety_function(function)
                return function
        raise ValueError(f"All safety functions of type {function_type} are already mapped")

    def unmap_safety_function(self, safe_func: "SafetyFunction") -> None:
        """Unmaps a safety function from the target map.

        Args:
            safe_func: The safety function to unmap.

        Raises:
            ValueError: if the safety function is not mapped.
        """
        map_input = [x.item for x in self.inputs]
        map_output = [x.item for x in self.outputs]
        io_unmapped = False
        for io in safe_func.ios.values():
            if isinstance(io, FSoEDictionaryItemInput) and io in map_input:
                self.inputs.transform_to_padding(io)
                io_unmapped = True
            elif isinstance(io, FSoEDictionaryItemOutput) and io in map_output:
                self.outputs.transform_to_padding(io)
                io_unmapped = True
            elif isinstance(io, FSoEDictionaryItemInputOutput):
                if io in map_input:
                    self.inputs.transform_to_padding(io)
                    io_unmapped = True
                if io in map_output:
                    self.outputs.transform_to_padding(io)
                    io_unmapped = True
        if not io_unmapped:
            self.logger.warning("The safety function is not mapped")

    def remove_safety_functions_by_type(
        self, handler: "FSoEMasterHandler", function_type: type["SafetyFunction"]
    ) -> "SafetyFunction":
        """Remove the last instance mapped of the given safety function type on the maps.

        Args:
            handler: FSoE handler.
            function_type: Safety function type.

        Returns:
            The safety function that was removed.

        Raises:
            ValueError: if no safety functions of the given type are found.
            ValueError: if no safety functions of the given type are mapped.

        """
        sf_available: list[SafetyFunction] = handler.safety_functions_by_type().get(
            function_type, []
        )
        if not sf_available:
            raise ValueError(f"No safety functions of type {function_type} found")
        for function in sf_available[::-1]:
            if self.is_safety_function_mapped(function):
                self.unmap_safety_function(function)
                return function
        raise ValueError(f"No safety functions of type {function_type} are mapped")

    def is_safety_function_mapped(
        self, safety_function: "SafetyFunction", strict: bool = True
    ) -> bool:
        """Check if the safety function is mapped in the process image.

        If at least one output of the safety function is present in the process image,
        the function is considered mapped. If the safety function has no outputs,
        it is considered mapped if at least one of its inputs is present in the maps.

        If strict is False, any IO mapped will return True.

        Args:
            safety_function: SafetyFunction to check.
            strict: if False, any IO mapped will return True. Defaults to True.

        Returns:
            True if the safety function is mapped, False otherwise.
        """
        map_input = [x.item for x in self.inputs]
        map_output = [x.item for x in self.outputs]
        has_output = any(
            isinstance(io, (FSoEDictionaryItemOutput, FSoEDictionaryItemInputOutput))
            for io in safety_function.ios.values()
        )
        if has_output:
            for io in safety_function.ios.values():
                if (
                    isinstance(io, (FSoEDictionaryItemOutput, FSoEDictionaryItemInputOutput))
                    and io in map_output
                ):
                    return True
        if not has_output or not strict:
            for io in safety_function.ios.values():
                if (
                    isinstance(io, (FSoEDictionaryItemInput, FSoEDictionaryItemInputOutput))
                    and io in map_input
                ):
                    return True
        return False

    @classmethod
    def from_rpdo_tpdo(
        cls, rpdo: RPDOMap, tpdo: TPDOMap, dictionary: "FSoEDictionary"
    ) -> "ProcessImage":
        """Create a ProcessImage instance from the given RPDO and TPDO maps.

        Returns:
            ProcessImage instance with the RPDO and TPDO maps filled.
        """
        process_image = cls.empty(dictionary)
        cls.__fill_dictionary_map_from_pdo(rpdo, process_image.outputs)
        cls.__fill_dictionary_map_from_pdo(tpdo, process_image.inputs)
        return process_image

    def fill_rpdo_map(self, rpdo_map: RPDOMap, servo_dictionary: "Dictionary") -> None:
        """Fill the RPDOMap used for the Safety Master PDU."""
        self.outputs.complete_with_padding()
        FSoEFrame.fill_pdo_map(
            self.outputs,
            servo_dictionary=servo_dictionary,
            pdo_map=rpdo_map,
            pdo_item_type=RPDOMapItem,
            frame_elements=MASTER_FRAME_ELEMENTS,
        )

    def fill_tpdo_map(self, tpdo_map: TPDOMap, servo_dictionary: "Dictionary") -> None:
        """Fill the TPDOMap used for the Safety Slave PDU."""
        self.inputs.complete_with_padding()
        FSoEFrame.fill_pdo_map(
            self.inputs,
            servo_dictionary=servo_dictionary,
            pdo_map=tpdo_map,
            pdo_item_type=TPDOMapItem,
            frame_elements=SLAVE_FRAME_ELEMENTS,
        )

    @staticmethod
    def __get_safety_bytes_range_from_pdo_length(pdo_byte_length: int) -> tuple[int, ...]:
        """Get the bytes that belong to the safe data in a PDO map according to its length.

        Args:
            pdo_byte_length: byte length of the PDO map

        Raises:
            ValueError: if pdo_byte_lenght is less than 6

        Returns:
            Tuple containing all byte indexes of the PDO map that belong to safe data
        """
        if pdo_byte_length < 6:
            raise ValueError("pdo_lenght must be at least 6")
        elif pdo_byte_length == 6:
            # Shortest PDOMap is 6 bytes, containing only one data byte
            return (1,)
        else:
            # The total bytes of data is the Pdo map length
            # minus 1 byte for the command and 2 bytes for the connection ID
            # divided by 4, since each data slot has 2 bytes of data and 2 bytes of CRC
            total_data_slots = (pdo_byte_length - 3) // 4
            return tuple(
                byt
                for slot_i in range(total_data_slots)
                for byt in (1 + slot_i * 4, 2 + slot_i * 4)
            )

    @classmethod
    def __fill_dictionary_map_from_pdo(
        cls, pdo_map: PDOMap, dictionary_map: "FSoEDictionaryMap"
    ) -> None:
        """Fill the dictionary map with items from the given PDOMap."""
        valid_bits = tuple(
            valid_bit
            for valid_byte in cls.__get_safety_bytes_range_from_pdo_length(
                pdo_map.data_length_bytes
            )
            for valid_bit in range(valid_byte * 8, (valid_byte + 1) * 8)
        )
        dictionary: FSoEDictionary = dictionary_map.dictionary
        position_bits = 0

        # Number of bits of variables that are in multiple data slots
        # That will be on future virtual padding
        pending_virtual_padding_bits = 0

        for pdo_item in pdo_map.items:
            if position_bits in valid_bits:
                # The item is a safe data item
                if pdo_item.register.idx == 0:
                    # Padding item

                    if pending_virtual_padding_bits == 0:
                        # Regular padding item
                        dictionary_map.add_padding(pdo_item.size_bits)
                    else:
                        # Virtual padding bits must not be added on the mapping,
                        # since they are already accounted on the fsoe map variable
                        if pending_virtual_padding_bits >= pdo_item.size_bits:
                            # There are not enough bits on this padding
                            # to fit all the remaining virtual ones
                            # Or is exactly the same size
                            pending_virtual_padding_bits -= pdo_item.size_bits
                        else:
                            # The padding is larger than the pending virtual bits
                            # All the virtual padding are already accounted,
                            # and some more must be added
                            dictionary_map.add_padding(
                                pdo_item.size_bits - pending_virtual_padding_bits
                            )
                            pending_virtual_padding_bits = 0

                else:
                    # Register item
                    register = pdo_item.register
                    fsoe_item = dictionary.name_map[register.identifier]
                    pending_virtual_padding_bits = fsoe_item.data_type.bits - pdo_item.size_bits
                    dictionary_map.add(fsoe_item)

            position_bits += pdo_item.size_bits

        dictionary_map.merge_adjacent_paddings()



import threading
from collections import OrderedDict
from collections.abc import Iterator
from random import randint
from typing import TYPE_CHECKING, Callable, Optional, TypeVar, Union, cast, overload

import ingenialogger
from ingenialink import RegDtype
from ingenialink.canopen.register import CanopenRegister
from ingenialink.dictionary import DictionarySafetyModule
from ingenialink.enums.register import RegAccess, RegCyclicType
from ingenialink.ethercat.servo import EthercatServo
from ingenialink.pdo import RPDOMap, TPDOMap
from ingenialink.utils._utils import convert_dtype_to_bytes

from ingeniamotion._utils import weak_lru
from ingeniamotion.enums import FSoEState
from ingeniamotion.exceptions import IMTimeoutError
from ingeniamotion.fsoe_master.fsoe import (
    FSOE_MASTER_INSTALLED,
    BaseMasterHandler,
    FSoEApplicationParameter,
    FSoEDataType,
    FSoEDictionary,
    FSoEDictionaryItem,
    FSoEDictionaryItemInput,
    FSoEDictionaryItemInputOutput,
    FSoEDictionaryItemOutput,
    State,
    StateData,
    calculate_sra_crc,
)
from ingeniamotion.fsoe_master.parameters import (
    PARAM_VALUE_TYPE,
    SafetyParameter,
    SafetyParameterDirectValidation,
)
from ingeniamotion.fsoe_master.process_image import ProcessImage
from ingeniamotion.fsoe_master.safety_functions import (
    SafeInputsFunction,
    SafetyFunction,
    SOutFunction,
    SS1Function,
    STOFunction,
)

if TYPE_CHECKING:
    from ingenialink.ethercat.dictionary import EthercatDictionary
    from ingenialink.ethercat.network import EthercatNetwork

SAFE_INSTANCE_TYPE = TypeVar("SAFE_INSTANCE_TYPE", bound="SafetyFunction")

__all__ = ["FSoEMasterHandler"]


class FSoEMasterHandler:
    """FSoE Master Handler.

    Args:
        slave_address: The servo's FSoE address.
        connection_id: The FSoE connection ID.
        watchdog_timeout: The FSoE master watchdog timeout in seconds.

    """

    FSOE_MANUF_SAFETY_ADDRESS = "FSOE_MANUF_SAFETY_ADDRESS"
    FSOE_DICTIONARY_CATEGORY = "FSOE"
    MDP_CONFIGURED_MODULE_1 = "MDP_CONFIGURED_MODULE_1"

    __FSOE_RPDO_MAP_UID = "ETG_COMMS_RPDO_MAP256"
    __FSOE_TPDO_MAP_UID = "ETG_COMMS_TPDO_MAP256"
    __FSOE_SAFETY_PROJECT_CRC = "FSOE_SAFETY_PROJECT_CRC"

    DEFAULT_WATCHDOG_TIMEOUT_S = 1

    def __init__(
        self,
        servo: EthercatServo,
        net: "EthercatNetwork",
        *,
        use_sra: bool,
        slave_address: Optional[int] = None,
        connection_id: Optional[int] = None,
        watchdog_timeout: float = DEFAULT_WATCHDOG_TIMEOUT_S,
        report_error_callback: Callable[[str, str], None],
        state_change_callback: Optional[Callable[[FSoEState], None]] = None,
    ):
        if not FSOE_MASTER_INSTALLED:
            return
        self.logger = ingenialogger.get_logger(__name__)

        self.__state_change_callback = state_change_callback
        self.__servo = servo
        self.__net = net
        self.__running: bool = False
        self.__uses_sra: bool = use_sra
        self.__is_subscribed_to_process_data_events: bool = False

        self.net.pdo_manager.subscribe_to_exceptions(self._pdo_thread_exception_handler)

        self.__state_is_data = threading.Event()

        # The saco slave might take a while to answer with a valid command
        # During it's initialization it will respond with 0's, that are ignored
        # To avoid triggering additional errors
        self.__in_initial_reset = False

        # Parameters that are part of the system
        # UID as key
        self.safety_parameters = OrderedDict[str, SafetyParameter]()

        # Parameters that will be transmitted during the fsoe parameter state
        fsoe_application_parameters: list[FSoEApplicationParameter] = []

        # Set MDP module
        safety_module = self.__set_configured_module_ident_1()

        for app_parameter in safety_module.application_parameters:
            register = servo.dictionary.get_register(app_parameter.uid)

            if safety_module.uses_sra:
                sp = SafetyParameter(register, servo)
            else:
                sp = SafetyParameterDirectValidation(register, servo)
                fsoe_application_parameters.append(sp.fsoe_application_parameter)

            self.safety_parameters[app_parameter.uid] = sp

        # If SRA is used, use a single application parameter with CRC computation
        self._sra_fsoe_application_parameter: Optional[FSoEApplicationParameter] = None
        if self.__uses_sra:
            self._sra_fsoe_application_parameter = FSoEApplicationParameter(
                name="SRA_CRC",
                initial_value=self.get_application_parameters_sra_crc(),
                n_bytes=4,
            )
            fsoe_application_parameters.append(self._sra_fsoe_application_parameter)

        self.dictionary = self.create_safe_dictionary(servo.dictionary)

        self.safety_functions = tuple(SafetyFunction.for_handler(self))

        if connection_id is None:
            connection_id = randint(1, 0xFFFF)

        self.__master_map_object = self.__servo.dictionary.get_object(self.__FSOE_RPDO_MAP_UID, 1)
        self.__slave_map_object = self.__servo.dictionary.get_object(self.__FSOE_TPDO_MAP_UID, 1)

        self.__safety_master_pdu = servo.read_rpdo_map_from_slave(self.__master_map_object)
        self.__safety_slave_pdu = servo.read_tpdo_map_from_slave(self.__slave_map_object)

        map_editable = (self.__master_map_object.registers[0].access == RegAccess.RW) and (
            self.__slave_map_object.registers[0].access == RegAccess.RW
        )

        try:
            self.__process_image = ProcessImage.from_rpdo_tpdo(
                self.__safety_master_pdu,
                self.__safety_slave_pdu,
                dictionary=self.dictionary,
            )
        except Exception as e:
            self.logger.error(
                "Error creating FSoE Process Image from RPDO and TPDO on the drive. "
                "Falling back to a default map.",
                exc_info=e,
            )
            self.__process_image = ProcessImage.default(self.dictionary)

        if not map_editable:
            self.__process_image.inputs._lock()
            self.__process_image.outputs._lock()

        self._master_handler = BaseMasterHandler(
            dictionary=self.dictionary,
            slave_address=slave_address
            if slave_address is not None
            else self.get_safety_address_from_slave(),
            connection_id=connection_id,
            watchdog_timeout_s=watchdog_timeout,
            application_parameters=fsoe_application_parameters,
            report_error_callback=report_error_callback,
            state_change_callback=self.__internal_state_change_callback,
            dictionary_map_is_editable=map_editable,
        )

        # If anything else fails on the constructor, ensure the master handler is deleted
        try:
            if slave_address is not None:
                self.set_safety_address(slave_address)

            self.set_process_image(self.__process_image)
        except Exception as ex:
            self._master_handler.delete()
            raise ex

    def subscribe_to_process_data_events(self) -> None:
        """Subscribes to process data events.

        Subscription may happen when setting the PDO maps to slaves or if
        the master is started manually (after a stop usually).
        """
        if self.__is_subscribed_to_process_data_events:
            return
        self.safety_master_pdu_map.subscribe_to_process_data_event(self.get_request)
        self.safety_slave_pdu_map.subscribe_to_process_data_event(self.set_reply)
        self.__is_subscribed_to_process_data_events = True

    def _pdo_thread_exception_handler(self, exc: Exception) -> None:
        """Callback method for the PDO thread exceptions.

        If there is an exception in the PDO thread and the master was running,
        it should be stopped.

        Args:
            exc: The exception that occurred.
        """
        self.logger.error(
            f"An exception occurred during the PDO exchange: {exc}. FSoE Master will be stopped."
        )
        if self.running:
            self.stop()

    @property
    def net(self) -> "EthercatNetwork":
        """Returns the Ethercat network instance."""
        return self.__net

    def get_application_parameters_sra_crc(self) -> int:
        """Calculates SRA CRC for the application parameters.

        SRA calculation needs as input a list of uint16 values:
            * The safety parameters are aggregated into a single byte
              array according to their dtype.
            * The resulting array is split into chunks of uint16 data.

        Returns:
            sra crc.
        """
        data = bytearray()
        for param in self.safety_parameters.values():
            bytes_data = convert_dtype_to_bytes(data=param.get(), dtype=param.register.dtype)
            data.extend(bytes_data)

        # Pad if odd number of bytes
        if len(data) % 2 != 0:
            data.append(0)

        # Convert to list of uint16 values
        serialized_data = [
            int.from_bytes(data[i : i + 2], "little") for i in range(0, len(data), 2)
        ]

        return cast("int", calculate_sra_crc(serialized_data))

    def __set_configured_module_ident_1(self) -> DictionarySafetyModule:
        """Sets the configured Module Ident.

        Returns:
            Module Ident that has been set.

        Raises:
            RuntimeError: if module ident value to write can not be retrieved.
        """
        module_ident = None
        skip_application_parameter = DictionarySafetyModule.ApplicationParameter(
            uid=self.__FSOE_SAFETY_PROJECT_CRC
        )
        for safety_module in self.__servo.dictionary.safety_modules.values():
            # FSOE_SAFETY_PROJECT_CRC should only be used by masters that do
            # not support SRA calculation
            if skip_application_parameter in safety_module.application_parameters:
                self.logger.warning(
                    "Safety module has the application parameter "
                    f"{self.__FSOE_SAFETY_PROJECT_CRC}, skipping it."
                )
                continue
            if self.__uses_sra and safety_module.uses_sra:
                module_ident = safety_module.module_ident
            if not self.__uses_sra and not safety_module.uses_sra:
                module_ident = safety_module.module_ident
            if module_ident is not None:
                break
        if module_ident is None:
            raise RuntimeError("Module ident value to write could not be retrieved.")

        self.__servo.write(self.MDP_CONFIGURED_MODULE_1, data=module_ident, subnode=0)

        return self.__servo.dictionary.get_safety_module(module_ident=module_ident)

    def __get_configured_module_ident_1(self) -> Union[int, float, str, bytes]:
        """Gets the configured Module Ident 1.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            Configured Module Ident 1.
        """
        return self.__servo.read(self.MDP_CONFIGURED_MODULE_1, subnode=0)

    def __get_safety_module(self) -> DictionarySafetyModule:
        """Gets the configured Module Ident 1.

        Args:
            servo: servo alias to reference it. ``default`` by default.

        Returns:
            Safety module.

        Raises:
            NotImplementedError: if the safety module uses SRA.
        """
        module_ident = int(self.__get_configured_module_ident_1())
        safety_module = self.__servo.dictionary.get_safety_module(module_ident=module_ident)
        if safety_module.uses_sra:
            self.logger.warning("Safety module with SRA is not available.")
        return safety_module

    def start(self) -> None:
        """Start the FSoE Master handler.

        Raises:
            RuntimeError: If the FSoE Master is already running.
        """
        if self.running:
            raise RuntimeError("FSoE Master is already running.")
        self.__in_initial_reset = True
        # Recalculate the SRA crc in case it changed
        if self._sra_fsoe_application_parameter is not None:
            self._sra_fsoe_application_parameter.set(self.get_application_parameters_sra_crc())
        self._master_handler.start()
        self.__running = True
        self.subscribe_to_process_data_events()

    def stop(self) -> None:
        """Stop the master handler."""
        self._master_handler.stop()
        self.__in_initial_reset = False
        self.__running = False

        self.safety_master_pdu_map.unsubscribe_to_process_data_event()
        self.safety_slave_pdu_map.unsubscribe_to_process_data_event()
        self.__is_subscribed_to_process_data_events = False

    def delete(self) -> None:
        """Delete the master handler."""
        self._master_handler.delete()

    @property
    def process_image(self) -> "ProcessImage":
        """Get the Process Image used for the Safety PDUs."""
        return self.__process_image

    def set_process_image(self, process_image: "ProcessImage") -> None:
        """Set new Process Image for the Safety PDUs.

        Raises:
            RuntimeError: If the FSoE Master is running.
        """
        if self.__running:
            raise RuntimeError("Cannot set map while the FSoE Master is running")

        self._master_handler.master.dictionary_map = process_image.outputs
        self._master_handler.slave.dictionary_map = process_image.inputs
        self.__process_image = process_image

    def configure_pdo_maps(self) -> None:
        """Configure the PDOMaps used for the Safety PDUs according to the process image.

        Raises:
            ValueError: If a register in the PDOMap has no identifier.
        """
        if self.__process_image.editable:
            # Fill the RPDOMap and TPDOMap with the items from the maps
            self.__process_image.fill_rpdo_map(self.safety_master_pdu_map, self.__servo.dictionary)
            self.__process_image.fill_tpdo_map(self.safety_slave_pdu_map, self.__servo.dictionary)

        # Initialize the safety master PDU map with zeros so that it has bytes set
        # without needing to enter get_request for it to have bytes set
        zero_bytes = bytes(self.safety_master_pdu_map.data_length_bytes)
        self.safety_master_pdu_map.set_item_bytes(zero_bytes)

        # Update the pdo maps elements that are safe parameters
        for pdu_map in (self.safety_master_pdu_map, self.safety_slave_pdu_map):
            for register, mapping_value in pdu_map.map_register_values().items():
                if register.identifier is None:
                    raise ValueError("Register in PDOMap has no identifier")
                if register.identifier in self.safety_parameters:
                    if mapping_value is None:
                        # Set parameter to zero if it is not mapped
                        # Although fw should ignore this parameter,
                        # it is better to reduce noise associated to it
                        mapping_value = 0
                    self.safety_parameters[register.identifier].set_without_updating(mapping_value)

    def set_pdo_maps_to_slave(self) -> None:
        """Set the PDOMaps to be used by the Safety PDUs to the slave."""
        # This function only assigns but does not update the values of the PDOMaps.
        # https://novantamotion.atlassian.net/browse/INGK-1140
        self.__servo.set_pdo_map_to_slave(
            rpdo_maps=[self.safety_master_pdu_map], tpdo_maps=[self.safety_slave_pdu_map]
        )

        # Subscribe to process data events
        self.subscribe_to_process_data_events()

        if self.__process_image.editable:
            self.safety_master_pdu_map.write_to_slave(padding=True)
            self.safety_slave_pdu_map.write_to_slave(padding=True)

    def remove_pdo_maps_from_slave(self) -> None:
        """Remove the PDOMaps used by the Safety PDUs from the slave."""
        self.__servo.remove_rpdo_map(self.safety_master_pdu_map)
        self.__servo.remove_tpdo_map(self.safety_slave_pdu_map)

    def get_request(self) -> None:
        """Set the FSoE master handler request to the Safety Master PDU PDOMap.

        Raises:
            RuntimeError: If the FSoE Master is not running.
        """
        if not self.__running:
            raise RuntimeError("FSoE Master is not running")
        req = self._master_handler.get_request()
        self.safety_master_pdu_map.set_item_bytes(req)

    def set_reply(self) -> None:
        """Get the FSoE slave response.

        It is extracted from the Safety Slave PDU PDOMap and set to the FSoE master handler.
        """
        reply = self.safety_slave_pdu_map.get_item_bytes()
        if self.__in_initial_reset:
            if reply[0] == 0:
                # Byte 0 of FSoE frame should always be the command
                # 0 is not a valid command
                return
            else:
                self.__in_initial_reset = False

        self._master_handler.set_reply(reply)

    def get_mismatched_parameters(
        self,
    ) -> Iterator[tuple[SafetyParameter, PARAM_VALUE_TYPE, PARAM_VALUE_TYPE]]:
        """Get parameters that are mismatched between the master and the slave.

        Additional method, with more contextual information than SRA or direct validation.
        to check parameters that are mismatched between the master and the slave.

        Yields:
            Mismatched parameters as tuples of (SafetyParameter, master_value, slave_value).
        """
        for param in self.safety_parameters.values():
            mismatched, master_value, slave_value = param.is_mismatched()
            if mismatched:
                yield param, master_value, slave_value

    def write_safe_parameters(self) -> None:
        """Write the safety parameters to the FSoE master handler.

        Warnings:
            PDO Maps that are safe parameters are not written here.
            They are configured during configure_pdo_maps.
        """
        pdu_map_registers = [
            *self.__master_map_object.registers,
            *self.__slave_map_object.registers,
        ]
        for param in self.safety_parameters.values():
            if param.register in pdu_map_registers:
                # Are configured during configure_pdo_maps
                continue
            param.set_to_slave()

    def get_parameters_not_related_to_safety_functions(self) -> set[SafetyParameter]:
        """Get the safety parameters that are not related to any safety function.

        Returns:
            The set of safety parameters that are not directly related to any safety function.
        """
        params = set(self.safety_parameters.values())
        for func in self.safety_functions:
            for param in func.parameters.values():
                if param in params:
                    params.remove(param)

        return params

    @weak_lru()
    def safety_functions_by_type(self) -> dict[type[SafetyFunction], list[SafetyFunction]]:
        """Get a dictionary with the safety functions grouped by type.

        Returns:
            A dictionary where the keys are the types of safety functions
                and the values are lists of instances of that type.
        """
        return {
            type(sf): [
                sf_of_type
                for sf_of_type in self.safety_functions
                if isinstance(sf_of_type, type(sf))
            ]
            for sf in self.safety_functions
        }

    @overload  # type: ignore[misc]
    def get_function_instance(self, typ: type[SAFE_INSTANCE_TYPE]) -> SAFE_INSTANCE_TYPE: ...

    @weak_lru()
    def get_function_instance(
        self, typ: type[SAFE_INSTANCE_TYPE], instance: Optional[int] = None
    ) -> SAFE_INSTANCE_TYPE:
        """Get the instance of a safety function.

        Raises:
            IndexError: If the instance index is out of range of the available instances.
            ValueError: If multiple instances of the type are found and
                no instance index is specified.
            ValueError: If no instance of the type is found.

        Args:
            typ: The type of the safety function to get.
            instance: The index of the instance to get.
                If None, if there's a single instance, it returns it.

        Returns:
            The instance of the safety function of the specified type.
        """
        funcs = [func for func in self.safety_functions if isinstance(func, typ)]

        if isinstance(instance, int):
            # First instance is 1
            index = instance - 1
            if index < 0 or index >= len(funcs):
                raise IndexError(
                    f"Master handler does not contain {typ.__name__} instance {instance}"
                )
            return funcs[index]
        elif len(funcs) == 0:
            raise ValueError(f"Master handler does not contain any {typ.__name__} instance")
        elif len(funcs) != 1:
            raise ValueError(
                f"Multiple {typ.__name__} instances found ({len(funcs)}). "
                f"Specify the instance number."
            )
        return funcs[0]

    @weak_lru()
    def sto_function(self) -> STOFunction:
        """Get the Safe Torque Off function.

        Returns:
            The Safe Torque Off function instance.
        """
        return self.get_function_instance(STOFunction)

    @weak_lru()
    def ss1_function(self) -> SS1Function:
        """Get the Safe Stop 1 function.

        Returns:
            The Safe Stop 1 function instance.
        """
        return self.get_function_instance(SS1Function)

    @weak_lru()
    def sout_function(self) -> Optional[SOutFunction]:
        """Get the Safe Output function.

        Returns:
            The Safe Output function instance.
            If there is no Safe Output function, None is returned.
        """
        try:
            return self.get_function_instance(SOutFunction)
        except Exception:
            return None

    @weak_lru()
    def safe_inputs_function(self) -> SafeInputsFunction:
        """Get the Safe Inputs function.

        Returns:
            The Safe Inputs function instance.
        """
        return self.get_function_instance(SafeInputsFunction)

    def set_fail_safe(self, fail_safe: bool) -> None:
        """Set the fail-safe mode of the FSoE master handler.

        When is on data state and fail-safe is active safe outputs will be sent but slaves
        should not take action.

        By standard, the fail-safe mode is activated when the master is not in Data state.

        Args:
            fail_safe: If True, the fail-safe mode is activated.
                If False, the fail-safe mode is deactivated.
        """
        self._master_handler.set_fail_safe(fail_safe)

    def sto_deactivate(self) -> None:
        """Set the STO command to deactivate the STO."""
        self.set_fail_safe(False)
        self.sto_function().command.set(True)

    def sto_activate(self) -> None:
        """Set the STO command to activate the STO."""
        self.sto_function().command.set(False)

    def ss1_deactivate(self) -> None:
        """Set the SS1 command to deactivate the SS1."""
        self.set_fail_safe(False)
        self.ss1_function().command.set(True)

    def ss1_activate(self) -> None:
        """Set the SS1 command to activate the SS1."""
        self.ss1_function().command.set(False)

    def sout_disable(self) -> None:
        """Deactivates SOUT.

        Raises:
            RuntimeError: If SOUT is not available.
        """
        sout_function: SOutFunction = self.sout_function()
        if sout_function is None:
            raise RuntimeError("SOUT not available.")
        sout_function.sout_disable.set(1)

    def sout_enable(self) -> None:
        """Activates SOUT.

        Raises:
            RuntimeError: If SOUT is not available.
        """
        sout_function: SOutFunction = self.sout_function()
        if sout_function is None:
            raise RuntimeError("SOUT not available.")
        sout_function.sout_disable.set(0)

    def safe_inputs_value(self) -> bool:
        """Get the safe inputs register value.

        Raises:
            ValueError: On unexpected value type.

        Returns:
            The safe inputs value as a boolean.
        """
        safe_inputs_value = self.safe_inputs_function().value.get()
        if not isinstance(safe_inputs_value, bool):
            raise ValueError(f"Wrong value type. Expected type bool, got {type(safe_inputs_value)}")
        return safe_inputs_value

    def get_safety_address(self) -> int:
        """Get the FSoE slave address configured on the master.

        Returns:
            The FSoE slave address.
        """
        return cast("int", self._master_handler.get_slave_address())

    def set_safety_address(self, address: int) -> None:
        """Set the drive's FSoE slave address to the master and the slave.

        Args:
            address: The address to be set.
        """
        self.__servo.write(self.FSOE_MANUF_SAFETY_ADDRESS, address)
        self._master_handler.set_slave_address(address)

    def get_safety_address_from_slave(self) -> int:
        """Get the FSoE slave address configured on the slave.

        Returns:
            The FSoE slave address.
        """
        return cast("int", self.__servo.read(self.FSOE_MANUF_SAFETY_ADDRESS))

    def is_sto_active(self) -> bool:
        """Check the STO state.

        Raises:
            ValueError: On unexpected value type.

        Returns:
            True if the STO is active. False otherwise.

        """
        sto_command = self.sto_function().command.get()
        if not isinstance(sto_command, bool):
            raise ValueError(f"Wrong value type. Expected type bool, got {type(sto_command)}")
        return sto_command

    def __internal_state_change_callback(self, state: "State") -> None:
        if state == StateData:
            self.__state_is_data.set()
        else:
            self.__state_is_data.clear()

        if self.__state_change_callback:
            self.__state_change_callback(FSoEState(state.id))

    def wait_for_data_state(self, timeout: Optional[float] = None) -> None:
        """Wait the FSoE master handler to reach the Data state.

        Args:
            timeout : how many seconds to wait for the FSoE master to reach the
                Data state, if ``None`` it will wait forever.
                ``None`` by default.

        Raises:
            IMTimeoutError: If the Data state is not reached within the timeout.

        """
        if self.__state_is_data.wait(timeout=timeout) is False:
            raise IMTimeoutError("The FSoE Master did not reach the Data state")

    @classmethod
    def create_safe_dictionary(cls, dictionary: "EthercatDictionary") -> "FSoEDictionary":
        """Create a FSoEdictionary with the safe inputs and outputs.

        Creates the FSoE dictionary from the servo's dictionary.

        Raises:
            TypeError: If the register is not of type CanopenRegister.

        Returns:
            A Dictionary instance with the safe inputs and outputs.

        """
        items = []
        for register in dictionary.registers(subnode=1).values():
            if register.cat_id != cls.FSOE_DICTIONARY_CATEGORY:
                continue

            if not isinstance(register, CanopenRegister):
                # Type could be narrowed to EthercatRegister
                # After this bugfix:
                # https://novantamotion.atlassian.net/browse/INGK-1111
                raise TypeError

            identifier = register.identifier
            if identifier is None:
                continue

            if identifier.startswith(("FSOE_SLAVE_FRAME", "FSOE_MASTER_FRAME")):
                # Elements of the standard FSoE frame are not added to the safe data dictionary
                continue

            if register.pdo_access in (
                RegCyclicType.SAFETY_OUTPUT,
                RegCyclicType.SAFETY_INPUT_OUTPUT,
                RegCyclicType.SAFETY_INPUT,
            ):
                items.append(cls._create_fsoe_dict_item_from_reg(register))

        return FSoEDictionary(items)

    @classmethod
    def _create_fsoe_dict_item_from_reg(cls, reg: "CanopenRegister") -> "FSoEDictionaryItem":
        # Create an arbitrary unique numeric key for the item
        # https://novantamotion.atlassian.net/browse/INGK-1112
        key = reg.idx * 1000 + reg.subidx

        data_typ = cls._reg_dtype_to_fsoe_data_type(reg.dtype)

        if reg.pdo_access == RegCyclicType.SAFETY_INPUT:
            return FSoEDictionaryItemInput(
                key,
                name=reg.identifier,
                data_type=data_typ,
                fail_safe_value=cls.__fsoe_input_data_type_default(
                    data_typ, FSoEDictionaryItemInput
                ),
            )
        elif reg.pdo_access == RegCyclicType.SAFETY_INPUT_OUTPUT:
            return FSoEDictionaryItemInputOutput(
                key,
                name=reg.identifier,
                data_type=data_typ,
                fail_safe_input_value=cls.__fsoe_input_data_type_default(
                    data_typ, FSoEDictionaryItemInputOutput
                ),
            )
        elif reg.pdo_access == RegCyclicType.SAFETY_OUTPUT:
            return FSoEDictionaryItemOutput(
                key,
                name=reg.identifier,
                data_type=data_typ,
            )

        raise NotImplementedError

    @classmethod
    def _reg_dtype_to_fsoe_data_type(cls, typ: "RegDtype") -> "FSoEDataType":
        if typ == RegDtype.U8:
            return FSoEDataType.UINT8
        if typ == RegDtype.U16:
            return FSoEDataType.UINT16
        if typ == RegDtype.U32:
            return FSoEDataType.UINT32
        if typ == RegDtype.S8:
            return FSoEDataType.INT8
        if typ == RegDtype.S16:
            return FSoEDataType.INT16
        if typ == RegDtype.S32:
            return FSoEDataType.INT32
        if typ == RegDtype.FLOAT:
            return FSoEDataType.FLOAT
        if typ == RegDtype.BOOL:
            return FSoEDataType.BOOL

        raise NotImplementedError(f"Unsupported register data type for FSoE: {typ}")

    @classmethod
    def __fsoe_input_data_type_default(
        cls, data_typ: "FSoEDataType", item_type: type["FSoEDictionaryItem"]
    ) -> Union[int, float, str, bytes]:
        if data_typ == FSoEDataType.BOOL:
            if item_type == FSoEDictionaryItemInput:
                # Inputs are assumed to be Low on safe-state
                return False
            if item_type == FSoEDictionaryItemInputOutput:
                # Input-Outputs are typically safe commands,
                # whose safe-state is being active
                return True

        if data_typ in (
            FSoEDataType.UINT8,
            FSoEDataType.UINT16,
            FSoEDataType.UINT32,
            FSoEDataType.INT8,
            FSoEDataType.INT16,
            FSoEDataType.INT32,
            FSoEDataType.FLOAT,
        ):
            return 0

        raise NotImplementedError(f"Unsupported data type for FSoE: {data_typ}")

    @property
    def safety_master_pdu_map(self) -> RPDOMap:
        """The PDOMap used for the Safety Master PDU.

        The PDOMap might not be up to date, call configure_pdo_maps first
        """
        return self.__safety_master_pdu

    @property
    def safety_slave_pdu_map(self) -> TPDOMap:
        """The PDOMap used for the Safety Slave PDU.

        The PDOMap might not be up to date, call configure_pdo_maps first
        """
        return self.__safety_slave_pdu

    @property
    def state(self) -> FSoEState:
        """Get the FSoE master state."""
        return FSoEState(self._master_handler.state.id)

    @property
    def running(self) -> bool:
        """True if FSoE Master is started, else False."""
        return self.__running
